@@ -56,7 +56,7 @@ Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  * Base type of Z80 controllers
  */
 
-cl_z80::cl_z80(int Itype, int Itech, class cl_sim *asim):
+cl_z80::cl_z80(struct cpu_entry *Itype, class cl_sim *asim):
   cl_uc(asim)
 {
   type= Itype;
@@ -65,11 +65,12 @@ cl_z80::cl_z80(int Itype, int Itech, class cl_sim *asim):
 int
 cl_z80::init(void)
 {
+
   cl_uc::init(); /* Memories now exist */
 
-  rom= address_space(MEM_ROM_ID);
+  //rom= address_space(MEM_ROM_ID);
 //  ram= mem(MEM_XRAM);
-  ram= rom;
+  //ram= rom;
 
   // zero out ram(this is assumed in regression tests)
   for (int i=0x8000; i<0x10000; i++) {
@@ -79,10 +80,10 @@ cl_z80::init(void)
   return(0);
 }
 
-const char *
+char *
 cl_z80::id_string(void)
 {
-  return("unspecified Z80");
+  return((char*)"unspecified Z80");
 }
 
 
@@ -107,7 +108,7 @@ void
 cl_z80::mk_hw_elements(void)
 {
   //class cl_base *o;
-  /* t_uc::mk_hw() does nothing */
+  cl_uc::mk_hw_elements();
 }
 
 void
@@ -115,7 +116,7 @@ cl_z80::make_memories(void)
 {
   class cl_address_space *as;
 
-  as= new cl_address_space("rom", 0, 0x10000, 8);
+  rom= ram= as= new cl_address_space("rom", 0, 0x10000, 8);
   as->init();
   address_spaces->add(as);
 
@@ -129,6 +130,121 @@ cl_z80::make_memories(void)
   ad->init();
   as->decoders->add(ad);
   ad->activate(0);
+
+  inputs= new cl_address_space("inputs", 0, 0x10000, 8);
+  inputs->init();
+  chip= new cl_memory_chip("in_chip", 0x10000, 8);
+  chip->init();
+  memchips->add(chip);
+  ad= new cl_address_decoder(inputs, chip, 0, 0xffff, 0);
+  ad->init();
+  inputs->decoders->add(ad);
+  address_spaces->add(inputs);
+  outputs= new cl_address_space("outputs", 0, 0x10000, 8);
+  outputs->init();
+  chip= new cl_memory_chip("out_chip", 0x10000, 8);
+  chip->init();
+  memchips->add(chip);
+  ad= new cl_address_decoder(outputs, chip, 0, 0xffff, 0);
+  ad->init();
+  outputs->decoders->add(ad);
+  address_spaces->add(outputs);
+  
+  regs8= new cl_address_space("regs8", 0, 16, 8);
+  regs8->init();
+  regs8->get_cell(0)->decode((t_mem*)&regs.raf.A);
+  regs8->get_cell(1)->decode((t_mem*)&regs.raf.F);
+  regs8->get_cell(2)->decode((t_mem*)&regs.bc.h);
+  regs8->get_cell(3)->decode((t_mem*)&regs.bc.l);
+  regs8->get_cell(4)->decode((t_mem*)&regs.de.h);
+  regs8->get_cell(5)->decode((t_mem*)&regs.de.l);
+  regs8->get_cell(6)->decode((t_mem*)&regs.hl.h);
+  regs8->get_cell(7)->decode((t_mem*)&regs.hl.l);
+
+  regs8->get_cell(8)->decode((t_mem*)&regs.ralt_af.aA);
+  regs8->get_cell(9)->decode((t_mem*)&regs.ralt_af.aF);
+  regs8->get_cell(10)->decode((t_mem*)&regs.a_bc.h);
+  regs8->get_cell(11)->decode((t_mem*)&regs.a_bc.l);
+  regs8->get_cell(12)->decode((t_mem*)&regs.a_de.h);
+  regs8->get_cell(13)->decode((t_mem*)&regs.a_de.l);
+  regs8->get_cell(14)->decode((t_mem*)&regs.a_hl.h);
+  regs8->get_cell(15)->decode((t_mem*)&regs.a_hl.l);
+
+  regs16= new cl_address_space("regs16", 0, 11, 16);
+  regs16->init();
+
+  regs16->get_cell(0)->decode((t_mem*)&regs.AF);
+  regs16->get_cell(1)->decode((t_mem*)&regs.BC);
+  regs16->get_cell(2)->decode((t_mem*)&regs.DE);
+  regs16->get_cell(3)->decode((t_mem*)&regs.HL);
+  regs16->get_cell(4)->decode((t_mem*)&regs.IX);
+  regs16->get_cell(5)->decode((t_mem*)&regs.IY);
+  regs16->get_cell(6)->decode((t_mem*)&regs.SP);
+  regs16->get_cell(7)->decode((t_mem*)&regs.aAF);
+  regs16->get_cell(8)->decode((t_mem*)&regs.aBC);
+  regs16->get_cell(9)->decode((t_mem*)&regs.aDE);
+  regs16->get_cell(10)->decode((t_mem*)&regs.aHL);
+
+  address_spaces->add(regs8);
+  address_spaces->add(regs16);
+
+  class cl_var *v;
+  vars->add(v= new cl_var(cchars("A"), regs8, 0, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("F"), regs8, 1, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("B"), regs8, 2, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("C"), regs8, 3, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("D"), regs8, 4, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("E"), regs8, 5, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("H"), regs8, 6, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("L"), regs8, 7, ""));
+  v->init();
+
+  vars->add(v= new cl_var(cchars("ALT_A"), regs8, 8, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_F"), regs8, 9, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_B"), regs8, 10, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_C"), regs8, 11, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_D"), regs8, 12, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_E"), regs8, 13, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_H"), regs8, 14, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_L"), regs8, 15, ""));
+  v->init();
+
+  vars->add(v= new cl_var(cchars("AF"), regs16, 0, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("BC"), regs16, 1, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("DE"), regs16, 2, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("HL"), regs16, 3, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("IX"), regs16, 4, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("IY"), regs16, 5, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("SP"), regs16, 6, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_AF"), regs16, 7, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_BC"), regs16, 8, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_DE"), regs16, 9, ""));
+  v->init();
+  vars->add(v= new cl_var(cchars("ALT_HL"), regs16, 10, ""));
+  v->init();
 }
 
 
@@ -207,12 +323,12 @@ cl_z80::get_disasm_info(t_addr addr,
   int start_addr = addr;
   struct dis_entry *dis_e;
 
-  code= get_mem(MEM_ROM_ID, addr++);
+  code= rom->get(addr++);
   dis_e = NULL;
 
   switch(code) {
     case 0xcb:  /* ESC code to lots of op-codes, all 2-byte */
-      code= get_mem(MEM_ROM_ID, addr++);
+      code= rom->get(addr++);
       i= 0;
       while ((code & disass_z80_cb[i].mask) != disass_z80_cb[i].code &&
         disass_z80_cb[i].mnemonic)
@@ -224,7 +340,7 @@ cl_z80::get_disasm_info(t_addr addr,
     break;
 
     case 0xed: /* ESC code to about 80 opcodes of various lengths */
-      code= get_mem(MEM_ROM_ID, addr++);
+      code= rom->get(addr++);
       i= 0;
       while ((code & disass_z80_ed[i].mask) != disass_z80_ed[i].code &&
         disass_z80_ed[i].mnemonic)
@@ -236,11 +352,11 @@ cl_z80::get_disasm_info(t_addr addr,
     break;
 
     case 0xdd: /* ESC codes,about 284, vary lengths, IX centric */
-      code= get_mem(MEM_ROM_ID, addr++);
+      code= rom->get(addr++);
       if (code == 0xcb) {
         immed_n = 2;
         addr++;  // pass up immed data
-        code= get_mem(MEM_ROM_ID, addr++);
+        code= rom->get(addr++);
         i= 0;
         while ((code & disass_z80_ddcb[i].mask) != disass_z80_ddcb[i].code &&
           disass_z80_ddcb[i].mnemonic)
@@ -262,11 +378,11 @@ cl_z80::get_disasm_info(t_addr addr,
     break;
 
     case 0xfd: /* ESC codes,sme as dd but IY centric */
-      code= get_mem(MEM_ROM_ID, addr++);
+      code= rom->get(addr++);
       if (code == 0xcb) {
         immed_n = 2;
         addr++;  // pass up immed data
-        code= get_mem(MEM_ROM_ID, addr++);
+        code= rom->get(addr++);
         i= 0;
         while ((code & disass_z80_fdcb[i].mask) != disass_z80_fdcb[i].code &&
           disass_z80_fdcb[i].mnemonic)
@@ -322,7 +438,7 @@ cl_z80::get_disasm_info(t_addr addr,
   return b;
 }
 
-const char *
+char *
 cl_z80::disass(t_addr addr, const char *sep)
 {
   char work[256], temp[20];
@@ -349,18 +465,18 @@ cl_z80::disass(t_addr addr, const char *sep)
           switch (*(b++))
             {
             case 'd': // d    jump relative target, signed? byte immediate operand
-              sprintf(temp, "#%d", (char)get_mem(MEM_ROM_ID, addr+immed_offset));
+              sprintf(temp, "#%d", (signed char)(rom->get(addr+immed_offset)));
               ++immed_offset;
               break;
             case 'w': // w    word immediate operand
               sprintf(temp, "#0x%04x",
-                 (uint)((get_mem(MEM_ROM_ID, addr+immed_offset)) |
-                        (get_mem(MEM_ROM_ID, addr+immed_offset+1)<<8)) );
+                 (uint)((rom->get(addr+immed_offset)) |
+                        (rom->get(addr+immed_offset+1)<<8)) );
               ++immed_offset;
               ++immed_offset;
               break;
             case 'b': // b    byte immediate operand
-              sprintf(temp, "#0x%02x", (uint)get_mem(MEM_ROM_ID, addr+immed_offset));
+              sprintf(temp, "#0x%02x", (uint)rom->get(addr+immed_offset));
               ++immed_offset;
               break;
             default:
@@ -406,16 +522,16 @@ void
 cl_z80::print_regs(class cl_console_base *con)
 {
   con->dd_printf("SZ-A-PNC  Flags= 0x%02x %3d %c  ",
-                 regs.F, regs.F, isprint(regs.F)?regs.F:'.');
+                 regs.raf.F, regs.raf.F, isprint(regs.raf.F)?regs.raf.F:'.');
   con->dd_printf("A= 0x%02x %3d %c\n",
-                 regs.A, regs.A, isprint(regs.A)?regs.A:'.');
+                 regs.raf.A, regs.raf.A, isprint(regs.raf.A)?regs.raf.A:'.');
   con->dd_printf("%c%c-%c-%c%c%c\n",
-                 (regs.F&BIT_S)?'1':'0',
-                 (regs.F&BIT_Z)?'1':'0',
-                 (regs.F&BIT_A)?'1':'0',
-                 (regs.F&BIT_P)?'1':'0',
-                 (regs.F&BIT_N)?'1':'0',
-                 (regs.F&BIT_C)?'1':'0');
+                 (regs.raf.F&BIT_S)?'1':'0',
+                 (regs.raf.F&BIT_Z)?'1':'0',
+                 (regs.raf.F&BIT_A)?'1':'0',
+                 (regs.raf.F&BIT_P)?'1':'0',
+                 (regs.raf.F&BIT_N)?'1':'0',
+                 (regs.raf.F&BIT_C)?'1':'0');
   con->dd_printf("BC= 0x%04x [BC]= %02x %3d %c  ",
                  regs.BC, ram->get(regs.BC), ram->get(regs.BC),
                  isprint(ram->get(regs.BC))?ram->get(regs.BC):'.');
@@ -581,7 +697,7 @@ cl_z80::exec_inst(void)
       /* DD escapes out to 2 to 4 byte opcodes(DD included)
         with a variety of uses.  It can precede the CB escape
         sequence to extend CB codes with IX+immed_byte */
-    case 0xdd: return(inst_dd());
+    case 0xdd: return(inst_dd(0xdd));
     case 0xde: return(inst_sbc(code));
     case 0xdf: return(inst_rst(code));
 
@@ -601,7 +717,7 @@ cl_z80::exec_inst(void)
     case 0xeb: return(inst_ex(code));
     case 0xec: return(inst_call(code));
       /* ED escapes out to misc IN, OUT and other oddball opcodes */
-    case 0xed: return(inst_ed());
+    case 0xed: return(inst_ed(0xed));
     case 0xee: return(inst_xor(code));
     case 0xef: return(inst_rst(code));
 
@@ -623,7 +739,7 @@ cl_z80::exec_inst(void)
       /* DD escapes out to 2 to 4 byte opcodes(DD included)
         with a variety of uses.  It can precede the CB escape
         sequence to extend CB codes with IX+immed_byte */
-    case 0xfd: return(inst_fd());
+    case 0xfd: return(inst_fd(0xfd));
     case 0xfe: return(inst_cp(code));
     case 0xff: return(inst_rst(code));
     }
@@ -638,24 +754,24 @@ cl_z80::exec_inst(void)
   return(resINV_INST);
 }
 
-void cl_z80::store1( TYPE_UWORD addr, t_mem val ) {
-  ram->set(addr, val);
+void cl_z80::store1( u16_t addr, t_mem val ) {
+  ram->write(addr, val);
 }
 
-void cl_z80::store2( TYPE_UWORD addr, TYPE_UWORD val ) {
-  ram->set(addr,   val & 0xff);
-  ram->set(addr+1, (val >> 8) & 0xff);
+void cl_z80::store2( u16_t addr, u16_t val ) {
+  ram->write(addr,   val & 0xff);
+  ram->write(addr+1, (val >> 8) & 0xff);
 }
 
-TYPE_UBYTE  cl_z80::get1( TYPE_UWORD addr ) {
-  return ram->get(addr);
+u8_t  cl_z80::get1( u16_t addr ) {
+  return ram->read(addr);
 }
 
-TYPE_UWORD  cl_z80::get2( TYPE_UWORD addr ) {
-  TYPE_UWORD  l, h;
+u16_t  cl_z80::get2( u16_t addr ) {
+  u16_t  l, h;
   
-  l = ram->get(addr  );
-  h = ram->get(addr+1);
+  l = ram->read(addr  );
+  h = ram->read(addr+1);
   
   return (h << 8) | l;
 }
@@ -664,8 +780,8 @@ t_mem       cl_z80::fetch1( void ) {
   return fetch( );
 }
 
-TYPE_UWORD  cl_z80::fetch2( void ) {
-  TYPE_UWORD  c1, c2;
+u16_t  cl_z80::fetch2( void ) {
+  u16_t  c1, c2;
   
   c1 = fetch( );
   c2 = fetch( );
@@ -676,17 +792,18 @@ t_mem       cl_z80::peek1 ( void ) {
   return rom->read(PC);
 }
 
-TYPE_UBYTE  cl_z80:: in_byte( TYPE_UWORD ioaddr )
+u8_t  cl_z80:: in_byte( u16_t ioaddr )
 {
-  return 0;
+  return inputs->read(ioaddr);
 }
 
-void        cl_z80::out_byte( TYPE_UWORD ioaddr, TYPE_UBYTE io_val )
+void        cl_z80::out_byte( u16_t ioaddr, u8_t io_val )
 {
+  outputs->write(ioaddr, io_val);
   return;
 }
 
-TYPE_UBYTE  cl_z80::reg_g_read ( t_mem g )
+u8_t  cl_z80::reg_g_read ( t_mem g )
 {
   switch( g )
     {
@@ -697,13 +814,13 @@ TYPE_UBYTE  cl_z80::reg_g_read ( t_mem g )
     case 4:  return regs.hl.h;
     case 5:  return regs.hl.l;
     case 6:  return get1( regs.HL );
-    case 7:  return regs.A;
+    case 7:  return regs.raf.A;
     default:
       return 0xffU;
     }
 }
 
-void        cl_z80::reg_g_store( t_mem g, TYPE_UBYTE new_val )
+void        cl_z80::reg_g_store( t_mem g, u8_t new_val )
 {
   switch( g )
     {
@@ -717,7 +834,7 @@ void        cl_z80::reg_g_store( t_mem g, TYPE_UBYTE new_val )
       store1( regs.HL, new_val );
       break;
 
-    case 7:  regs.A    = new_val;  break;  /* write to a */
+    case 7:  regs.raf.A    = new_val;  break;  /* write to a */
     }
 }
 
