@@ -19,6 +19,12 @@
 #include <stdio.h>
 #include <stdlib.h>
 
+#include "SDCCglobl.h"
+#ifdef HAVE_BACKTRACE_SYMBOLS_FD
+#include <unistd.h>
+#include <execinfo.h>
+#endif
+
 #include "SDCCerr.h"
 
 #define NELEM(x) (sizeof (x) / sizeof *(x))
@@ -54,7 +60,7 @@ struct
   { E_SYNTAX_ERROR, ERROR_LEVEL_ERROR,
      "Syntax error, declaration ignored at '%s'", 0 },
   { E_CONST_EXPECTED, ERROR_LEVEL_ERROR,
-      "Initializer element is not a compile-time constant", 0 },
+      "Initializer element is not a constant expression", 0 },
   { E_OUT_OF_MEM, ERROR_LEVEL_ERROR,
      "'malloc' failed file '%s' for size %ld", 0 },
   { E_FILE_OPEN_ERR, ERROR_LEVEL_ERROR,
@@ -111,10 +117,10 @@ struct
      "'unary %c': illegal operand", 0 },
   { E_CONV_ERR, ERROR_LEVEL_ERROR,
      "conversion error: integral promotion failed", 0 },
-  { E_INT_REQD, ERROR_LEVEL_ERROR,
-     "type must be INT for bit field definition", 0 },
+  { E_BITFLD_TYPE, ERROR_LEVEL_ERROR,
+     "invalid type for bit-field", 0 },
   { E_BITFLD_SIZE, ERROR_LEVEL_ERROR,
-     "bit field size cannot be greater than int (%d bits)", 0 },
+     "bit-field size too wide for type (max %d bits)", 0 },
   { W_TRUNCATION, ERROR_LEVEL_WARNING,
      "high order truncation might occur", 0 },
   { E_CODE_WRITE, ERROR_LEVEL_ERROR,
@@ -140,15 +146,15 @@ struct
   { E_SHIFT_OP_INVALID, ERROR_LEVEL_ERROR,
      "invalid operand for shift operator", 0 },
   { E_COMPARE_OP, ERROR_LEVEL_ERROR,
-     "compare operand cannot be struct/union", 0 },
+     "operands are not comparable", 0 },
   { E_BITWISE_OP, ERROR_LEVEL_ERROR,
      "operand invalid for bitwise operation", 0 },
   { E_ANDOR_OP, ERROR_LEVEL_ERROR,
      "Invalid operand for '&&' or '||'", 0 },
   { E_TYPE_MISMATCH, ERROR_LEVEL_ERROR,
      "indirections to different types %s %s ", 0 },
-  { E_AGGR_ASSIGN, ERROR_LEVEL_ERROR,
-     "cannot assign values to aggregates", 0 },
+  { E_ARRAY_ASSIGN, ERROR_LEVEL_ERROR,
+     "cannot assign values to arrays", 0 },
   { E_ARRAY_DIRECT, ERROR_LEVEL_ERROR,
      "bit Arrays can be accessed by literal index only", 0 },
   { E_BIT_ARRAY, ERROR_LEVEL_ERROR,
@@ -247,8 +253,8 @@ struct
      "Pre-Processor %s", 0 },
   { E_STRUCT_AS_ARG, ERROR_LEVEL_ERROR,
      "SDCC cannot pass structure '%s' as function argument", 0 },
-  { E_PREV_DEF_CONFLICT, ERROR_LEVEL_ERROR,
-     "conflict with previous definition of '%s' for attribute '%s'", 0 },
+  { E_PREV_DECL_CONFLICT, ERROR_LEVEL_ERROR,
+     "conflict with previous declaration of '%s' for attribute '%s' at %s:%d", 0 },
   { E_CODE_NO_INIT, ERROR_LEVEL_WARNING,
      "variable '%s' declared in code space must have initialiser", 0 },
   { E_OPS_INTEGRAL, ERROR_LEVEL_ERROR,
@@ -295,8 +301,8 @@ struct
      "cannot compile more than one source file. file '%s' ignored", 0 },
   { I_CYCLOMATIC, ERROR_LEVEL_INFO,
      "function '%s', # edges %d , # nodes %d , cyclomatic complexity %d", 0 },
-  { E_DIVIDE_BY_ZERO, ERROR_LEVEL_ERROR,
-     "dividing by ZERO", 0 },
+  { E_DIVIDE_BY_ZERO, ERROR_LEVEL_WARNING,
+     "dividing by 0", 0 },
   { E_FUNC_BIT, ERROR_LEVEL_ERROR,
      "function cannot return 'bit'", 0 },
   { E_CAST_ZERO, ERROR_LEVEL_ERROR,
@@ -413,7 +419,7 @@ struct
      "size of void is zero", 0 },
   { W_POSSBUG2, ERROR_LEVEL_WARNING,
      "possible code generation error at %s line %d,\n"
-     " please report problem and send source code at SDCC-USER list on SF.Net"},
+     " please report problem and send source code at sdcc-user list on sourceforge.net"},
   { W_COMPLEMENT, ERROR_LEVEL_WARNING,
      "using ~ on bit/bool/unsigned char variables can give unexpected results due to promotion to int", 0 },
   { E_SHADOWREGS_NO_ISR, ERROR_LEVEL_ERROR,
@@ -439,7 +445,7 @@ struct
   { W_BAD_PRAGMA_ARGUMENTS, ERROR_LEVEL_WARNING,
      "#pragma %s: bad argument(s); pragma ignored", 0 },
   { E_BAD_RESTRICT, ERROR_LEVEL_ERROR,
-     "Only pointers may be qualified with 'restrict'", 0 },
+     "Only object pointers may be qualified with 'restrict'", 0 },
   { E_BAD_INLINE, ERROR_LEVEL_ERROR,
      "Only functions may be qualified with 'inline'", 0 },
   { E_BAD_INT_ARGUMENT, ERROR_LEVEL_ERROR,
@@ -451,7 +457,7 @@ struct
   { W_DEPRECATED_KEYWORD, ERROR_LEVEL_WARNING,
      "keyword '%s' is deprecated, use '%s' instead", 0 },
   { E_STORAGE_CLASS_FOR_PARAMETER, ERROR_LEVEL_ERROR,
-     "storage class specified for parameter '%s'", 0 },
+     "storage class other than register specified for parameter '%s'", 0 },
   { E_OFFSETOF_TYPE, ERROR_LEVEL_ERROR,
      "offsetof can only be applied to structs/unions", 0 },
   { E_INCOMPLETE_FIELD, ERROR_LEVEL_ERROR,
@@ -526,13 +532,35 @@ struct
   { E_WCHAR_STRING_C11, ERROR_LEVEL_ERROR,
     "wide character string of type u8, u, U requires C11 or later", 0},
   { W_UNKNOWN_REG, ERROR_LEVEL_WARNING,
-    "unknown register specification", 0},
+    "unknown register specification %s", 0},
   { E_HEXFLOAT_C99, ERROR_LEVEL_ERROR,
     "hexadecimal floating constant requires ISO C99 or later", 0},
   { E_ANONYMOUS_STRUCT_TAG, ERROR_LEVEL_ERROR,
     "tagged anonymous struct/union '%s'", 0},
   { W_INLINE_FUNCATTR, ERROR_LEVEL_WARNING,
     "inline function '%s' might lose function attributes", 0},
+  { E_FOR_INITAL_DECLARATION_C99, ERROR_LEVEL_ERROR,
+    "initial declaration in for loop requires ISO C99 or later", 0},
+  { E_QUALIFIED_ARRAY_PARAM_C99, ERROR_LEVEL_ERROR,
+    "qualifiers in array parameters require ISO C99 or later", 0},
+  { E_QUALIFIED_ARRAY_NOPARAM, ERROR_LEVEL_ERROR,
+    "qualifier or static in array declarator that is not a parameter", 0},
+  { E_STATIC_ARRAY_PARAM_C99, ERROR_LEVEL_ERROR,
+    "static in array parameters requires ISO C99 or later", 0},
+  { E_INT_MULTIPLE, ERROR_LEVEL_ERROR,
+    "multiple interrupt numbers for '%s'", 0},
+  { W_INCOMPAT_PTYPES, ERROR_LEVEL_WARNING,
+     "pointer types incompatible ", 0 },
+  { E_STATIC_ASSERTION_C2X, ERROR_LEVEL_ERROR,
+    "static assertion with one argument requires C2X or later", 0 },
+  { W_STATIC_ASSERTION_2, ERROR_LEVEL_WARNING,
+    "static assertion failed", 0 },
+  { E_DECL_AFTER_STATEMENT_C99, ERROR_LEVEL_ERROR,
+    "declaration after statement requires ISO C99 or later", 0 },
+  { E_SHORTCALL_INVALID_VALUE, ERROR_LEVEL_ERROR,
+    "invalid value for __z88dk_shortcall %s parameter: %x", 0},
+  { E_DUPLICATE_PARAMTER_NAME, ERROR_LEVEL_ERROR,
+    "duplicate parameter name %s for function %s", 0},
 };
 
 /* -------------------------------------------------------------------------------
@@ -655,6 +683,29 @@ werror (int errNum, ...)
   return ret;
 }
 
+/* -------------------------------------------------------------------------------
+werror_bt - like werror(), but als provide a backtrace
+ * -------------------------------------------------------------------------------
+ */
+int
+werror_bt (int errNum, ...)
+{
+#ifdef HAVE_BACKTRACE_SYMBOLS_FD
+  void *callstack[16];
+  int frames = backtrace (callstack, 16);
+  fprintf (stderr, "Backtrace:\n");
+  backtrace_symbols_fd (callstack, frames, STDERR_FILENO);
+#endif
+
+  int ret;
+  va_list marker;
+  va_start (marker, errNum);
+  ret = vwerror (errNum, marker);
+  va_end (marker);
+
+  return ret;
+}
+
 /* -------------------------------------------------------------------------------
  * werrorfl - Output a standard error message with variable number of arguments.
  *            Use a specified filename and line number instead of the default.
@@ -717,6 +768,23 @@ setWarningDisabled (int errNum)
     ErrTab[errNum].disabled = 1;
 }
 
+/* -------------------------------------------------------------------------------
+ * disabledState - Enable/Disable output of specified warning
+ * -------------------------------------------------------------------------------
+ */
+int
+setWarningDisabledState (int errNum, int disabled)
+{
+  if ((errNum >= 0) && (errNum < NELEM (ErrTab)) && (ErrTab[errNum].errType <= ERROR_LEVEL_WARNING))
+  {
+    int originalState = ErrTab[errNum].disabled;
+    ErrTab[errNum].disabled = disabled;
+    return originalState;
+  }
+  return 0;
+}
+
+
 /* -------------------------------------------------------------------------------
  * Set the flag to treat warnings as errors
  * -------------------------------------------------------------------------------
