@@ -56,8 +56,8 @@ enum {
   E_PTR_REQD                    =  27, /* pointer required     */
   E_UNARY_OP                    =  28, /* unary operator bad op*/
   E_CONV_ERR                    =  29, /* conversion error     */
-  E_INT_REQD                    =  30, /* bit field must be int*/
-  E_BITFLD_SIZE                 =  31, /* bit field size > 16  */
+  E_BITFLD_TYPE                 =  30, /* invalid type for bit-field */
+  E_BITFLD_SIZE                 =  31, /* bit-field too wide for type */
   W_TRUNCATION                  =  32, /* high order trucation */
   E_CODE_WRITE                  =  33, /* trying 2 write to code */
   E_LVALUE_CONST                =  34, /* lvalue is a const   */
@@ -74,7 +74,7 @@ enum {
   E_BITWISE_OP                  =  45, /* bit op invalid op   */
   E_ANDOR_OP                    =  46, /* && || op invalid    */
   E_TYPE_MISMATCH               =  47, /* type mismatch       */
-  E_AGGR_ASSIGN                 =  48, /* aggr assign         */
+  E_ARRAY_ASSIGN                =  48, /* array assign        */
   E_ARRAY_DIRECT                =  49, /* array indexing in   */
   E_BIT_ARRAY                   =  50, /* bit array not allowed  */
   E_DUPLICATE_TYPEDEF           =  51, /* typedef name duplicate */
@@ -124,7 +124,7 @@ enum {
   W_FUNC_NO_RETURN              =  95, /* no return statement found */
   W_PRE_PROC_WARNING            =  96, /* preprocessor generated warning */
   E_STRUCT_AS_ARG               =  97, /* structure passed as argument */
-  E_PREV_DEF_CONFLICT           =  98, /* previous definition conflicts with current */
+  E_PREV_DECL_CONFLICT          =  98, /* previous declaration conflicts with current */
   E_CODE_NO_INIT                =  99, /* vars in code space must have initializer */
   E_OPS_INTEGRAL                = 100, /* operans must be integral for certain assignments */
   E_TOO_MANY_PARMS              = 101, /* too many parameters */
@@ -218,7 +218,7 @@ enum {
   E_FLEXARRAY_INEMPTYSTRCT      = 189, /* flexible array in otherwise empty struct */
   W_EMPTY_SOURCE_FILE           = 190, /* ISO C forbids an empty source file */
   W_BAD_PRAGMA_ARGUMENTS        = 191, /* #pragma %s: bad argument(s); pragma ignored */
-  E_BAD_RESTRICT                = 192, /* Only pointers may be qualified with 'restrict' */
+  E_BAD_RESTRICT                = 192, /* Only object pointers may be qualified with 'restrict' */
   E_BAD_INLINE                  = 193, /* Only functions may be qualified with 'inline' */
   E_BAD_INT_ARGUMENT            = 194, /* Bad integer option argument */
   E_NEGATIVE_ARRAY_SIZE         = 195, /* Size of array '%s' is negative */
@@ -265,6 +265,17 @@ enum {
   E_HEXFLOAT_C99                = 236, /* hexadecimal floating constant requires ISO C99 or later */
   E_ANONYMOUS_STRUCT_TAG        = 237, /* anonymous struct/union should not have a tag */
   W_INLINE_FUNCATTR             = 238, /* inline functions should not be z88dk_fastcall or z88dk_callee */
+  E_FOR_INITAL_DECLARATION_C99  = 239, /* initial declaration in for loop requires ISO C99 or later */
+  E_QUALIFIED_ARRAY_PARAM_C99   = 240, /* qualifiers in array parameters require ISO C99 or later */
+  E_QUALIFIED_ARRAY_NOPARAM     = 241, /* qualifier or static in array declarator that is not a parameter */
+  E_STATIC_ARRAY_PARAM_C99      = 242, /* static in array parameters requires ISO C99 or later */
+  E_INT_MULTIPLE                = 243, /* multiple interrupt numbers */
+  W_INCOMPAT_PTYPES             = 244, /* incompatible pointer assignment (not allowed by the standard, but allowed in SDCC) */
+  E_STATIC_ASSERTION_C2X        = 245, /* static assertion with one argument requires C2X or later */
+  W_STATIC_ASSERTION_2          = 246, /* static assertion failed */
+  E_DECL_AFTER_STATEMENT_C99    = 247, /* declaration after statement requires ISO C99 or later */
+  E_SHORTCALL_INVALID_VALUE     = 248, /* Invalid value for a __z88dk_shortcall specifier */
+  E_DUPLICATE_PARAMTER_NAME     = 249, /* duplicate parameter name */
 
   /* don't touch this! */
   NUMBER_OF_ERROR_MESSAGES             /* Number of error messages */
@@ -280,6 +291,11 @@ enum {
 # define assert(expr) ((expr) ? (void)0 : fatal (1, E_INTERNAL_ERROR, __FILE__, __LINE__, #expr))
 #endif
 
+#define wassertl_bt(a,s)   (void)((a) ? 0 : \
+        (werror_bt (E_INTERNAL_ERROR, __FILE__, __LINE__, s), 0))
+
+#define wassert_bt(a) wassertl_bt(a, "code generator internal error")
+
 /** Describes the maximum error level that will be logged.  Any level
  *  includes all of the levels listed after it.
  *
@@ -344,6 +360,15 @@ werror - Output a standard eror message with variable number of arguements
 
 int werror (int errNum, ... );
 
+/*
+-------------------------------------------------------------------------------
+werror_bt - like werror(), but als provide a backtrace
+
+-------------------------------------------------------------------------------
+*/
+
+int werror_bt (int errNum, ... );
+
 /*
 -------------------------------------------------------------------------------
 werrorfl - Output a standard eror message with variable number of arguements.
@@ -379,6 +404,14 @@ disabled - Disable output of specified warning
 
 void setWarningDisabled (int errNum);
 
+/*
+-------------------------------------------------------------------------------
+disabledState - Enable/Disable output of specified warning
+-------------------------------------------------------------------------------
+*/
+
+int setWarningDisabledState (int errNum, int disabled);
+
 /*
 -------------------------------------------------------------------------------
 Set the flag to treat warnings as errors
