@@ -65,8 +65,8 @@ aopLiteralGptr (const char * name, value * val)
   return dbuf_detach_c_str (&dbuf);
 }
 
-char *
-aopLiteralLong (value * val, int offset, int size)
+const char *
+aopLiteralLong (value *val, int offset, int size)
 {
   unsigned long v;
   struct dbuf_s dbuf;
@@ -107,8 +107,8 @@ aopLiteralLong (value * val, int offset, int size)
 /*-----------------------------------------------------------------*/
 /* aopLiteral - string from a literal value                        */
 /*-----------------------------------------------------------------*/
-char *
-aopLiteral (value * val, int offset)
+const char *
+aopLiteral (value *val, int offset)
 {
   return aopLiteralLong (val, offset, 1);
 }
@@ -131,14 +131,14 @@ emitDebugSym (struct dbuf_s *oBuf, symbol * sym)
     {
       dbuf_printf (oBuf, "G$");
     }
-  dbuf_printf (oBuf, "%s$%d$%d", sym->name, sym->level, sym->block);
+  dbuf_printf (oBuf, "%s$%ld_%ld$%d", sym->name, sym->level / LEVEL_UNIT, sym->level % LEVEL_UNIT, sym->block);
 }
 
 /*-----------------------------------------------------------------*/
 /* emitRegularMap - emit code for maps with no special cases       */
 /*-----------------------------------------------------------------*/
 static void
-emitRegularMap (memmap * map, bool addPublics, bool arFlag)
+emitRegularMap (memmap *map, bool addPublics, bool arFlag)
 {
   symbol *sym;
   ast *ival = NULL;
@@ -152,7 +152,11 @@ emitRegularMap (memmap * map, bool addPublics, bool arFlag)
       if (!strcmp (map->sname, CODE_NAME))
         dbuf_tprintf (&map->oBuf, "\t!areacode\n", map->sname);
       else if (!strcmp (map->sname, DATA_NAME))
-        dbuf_tprintf (&map->oBuf, "\t!areadata\n", map->sname);
+        {
+          dbuf_tprintf (&map->oBuf, "\t!areadata\n", map->sname);
+          if (options.data_seg && strcmp (DATA_NAME, options.data_seg))
+            dbuf_tprintf (&map->oBuf, "\t!area\n", options.data_seg);
+        }
       else if (!strcmp (map->sname, HOME_NAME))
         dbuf_tprintf (&map->oBuf, "\t!areahome\n", map->sname);
       else
@@ -341,17 +345,6 @@ emitRegularMap (memmap * map, bool addPublics, bool arFlag)
               emitDebugSym (&map->oBuf, sym);
               dbuf_printf (&map->oBuf, " == 0x%04x\n", SPEC_ADDR (sym->etype));
             }
-          if (TARGET_IS_XA51)
-            {
-              if (map == sfr)
-                {
-                  equ = "sfr";
-                }
-              else if (map == bit || map == sfrbit)
-                {
-                  equ = "bit";
-                }
-            }
           dbuf_printf (&map->oBuf, "%s\t%s\t0x%04x\n", sym->rname, equ, SPEC_ADDR (sym->etype));
         }
       else
@@ -387,7 +380,7 @@ emitRegularMap (memmap * map, bool addPublics, bool arFlag)
 /* initValPointer - pointer initialization code massaging          */
 /*-----------------------------------------------------------------*/
 value *
-initValPointer (ast * expr)
+initValPointer (ast *expr)
 {
   value *val;
 
@@ -535,7 +528,7 @@ initValPointer (ast * expr)
 /* initPointer - pointer initialization code massaging             */
 /*-----------------------------------------------------------------*/
 value *
-initPointer (initList * ilist, sym_link * toType, int showError)
+initPointer (initList *ilist, sym_link *toType, int showError)
 {
   value *val;
   ast *expr;
@@ -548,7 +541,11 @@ initPointer (initList * ilist, sym_link * toType, int showError)
   expr = list2expr (ilist);
 
   if (!expr)
-    goto wrong;
+    {
+      if (showError)
+        werror (E_CONST_EXPECTED);
+      return 0;
+    }
 
   /* try it the old way first */
   if ((val = constExprValue (expr, FALSE)))
@@ -567,7 +564,7 @@ initPointer (initList * ilist, sym_link * toType, int showError)
   /* (char *)(expr1) */
   if (IS_CAST_OP (expr))
     {
-      if (compareType (toType, expr->left->ftype) == 0)
+      if (compareType (toType, expr->left->ftype) == 0 && showError)
         {
           werror (W_INIT_WRONG);
           printFromToType (expr->left->ftype, toType);
@@ -590,13 +587,14 @@ initPointer (initList * ilist, sym_link * toType, int showError)
   if (val)
     return val;
 
-wrong:
   if (showError)
-    if (expr)
-      werrorfl (expr->filename, expr->lineno, E_INCOMPAT_PTYPES);
-    else
-      werror (E_INCOMPAT_PTYPES);
-  return NULL;
+    {
+      if (expr)
+        werrorfl (expr->filename, expr->lineno, E_CONST_EXPECTED);
+      else
+        werror (E_CONST_EXPECTED);
+    }
+  return 0;
 }
 
 /*-----------------------------------------------------------------*/
@@ -606,17 +604,26 @@ void
 printChar (struct dbuf_s *oBuf, const char *s, int plen)
 {
   int i;
-  int len = plen;
   int pplen = 0;
   char buf[100];
   char *p = buf;
-  int strEnd = plen - 1;
 
-  if (s)
-    while (s[strEnd] != 0)
-      strEnd--;
+  if (TARGET_PDK_LIKE && !TARGET_IS_PDK16) // Assembler does not support .ascii
+    {
+      while (pplen < plen)
+        {
+          if (isprint((unsigned char) *s))
+            dbuf_tprintf (oBuf, "\t!db !constbyte\t; %c\n", (unsigned char) *s, (unsigned char) *s);
+          else
+            dbuf_tprintf (oBuf, "\t!db !constbyte\n", (unsigned char) *s);
 
-  while (len && pplen < plen)
+          s++;
+          pplen++;
+        }
+      return;
+    }
+
+  while (pplen < plen)
     {
       i = 60;
       while (i && pplen < plen)
@@ -626,34 +633,26 @@ printChar (struct dbuf_s *oBuf, const char *s, int plen)
               *p = '\0';
               if (p != buf)
                 dbuf_tprintf (oBuf, "\t!ascii\n", buf);
-              dbuf_tprintf (oBuf, "\t!db !constbyte\n", pplen < strEnd ? ((unsigned char) *s) : 0x00);
+              dbuf_tprintf (oBuf, "\t!db !constbyte\n", (unsigned char) *s);
               p = buf;
+              i = 60;
             }
           else
             {
               *p = *s;
               p++;
+              i--;
             }
           s++;
           pplen++;
-          i--;
         }
       if (p != buf)
         {
           *p = '\0';
           dbuf_tprintf (oBuf, "\t!ascii\n", buf);
           p = buf;
+          i = 60;
         }
-
-      if (len > 60)
-        len -= 60;
-      else
-        len = 0;
-    }
-  while (pplen < plen)
-    {
-      dbuf_tprintf (oBuf, "\t!db !constbyte\n", 0);
-      pplen++;
     }
 }
 
@@ -664,15 +663,15 @@ void
 printChar16 (struct dbuf_s *oBuf, const TYPE_TARGET_UINT *s, int plen)
 {
   int pplen = 0;
-  int strEnd = plen - 1;
-
-  if (s)
-    while (s[strEnd] != 0)
-      strEnd--;
 
   while (pplen < plen)
     {
-      if (port->little_endian)
+      if (TARGET_PDK_LIKE && !TARGET_IS_PDK16)
+        {
+          dbuf_tprintf (oBuf, "\t!db !constbyte\n", (*s >> 0) & 0xff);
+          dbuf_tprintf (oBuf, "\t!db !constbyte\n", (*s >> 8) & 0xff);
+        }
+      else if (port->little_endian)
         dbuf_printf (oBuf, "\t.byte %d,%d\n", (*s >> 0) & 0xff, (*s >> 8) & 0xff);
       else
         dbuf_printf (oBuf, "\t.byte %d,%d\n", (*s >> 8) & 0xff, (*s >> 0) & 0xff);
@@ -694,15 +693,17 @@ void
 printChar32 (struct dbuf_s *oBuf, const TYPE_TARGET_ULONG *s, int plen)
 {
   int pplen = 0;
-  int strEnd = plen - 1;
-
-  if (s)
-    while (s[strEnd] != 0)
-      strEnd--;
 
   while (pplen < plen)
     {
-      if (port->little_endian)
+      if (TARGET_PDK_LIKE && !TARGET_IS_PDK16)
+        {
+          dbuf_tprintf (oBuf, "\t!db !constbyte\n", (*s >> 0) & 0xff);
+          dbuf_tprintf (oBuf, "\t!db !constbyte\n", (*s >> 8) & 0xff);
+          dbuf_tprintf (oBuf, "\t!db !constbyte\n", (*s >> 16) & 0xff);
+          dbuf_tprintf (oBuf, "\t!db !constbyte\n", (*s >> 24) & 0xff);
+        }
+      else if (port->little_endian)
         dbuf_printf (oBuf, "\t.byte %d,%d,%d,%d\n", (*s >> 0) & 0xff, (*s >> 8) & 0xff, (*s >> 16) & 0xff, (*s >> 24) & 0xff);
       else
         dbuf_printf (oBuf, "\t.byte %d,%d,%d,%d\n", (*s >> 24) & 0xff, (*s >> 16) & 0xff, (*s >> 8) & 0xff,(*s >> 0) & 0xff);
@@ -745,6 +746,35 @@ pointerTypeToGPByte (const int p_type, const char *iname, const char *oname)
   return -1;
 }
 
+/*-----------------------------------------------------------------*/
+/* printIvalVal - generate ival according from value               */
+/*-----------------------------------------------------------------*/
+static void printIvalVal(struct dbuf_s *oBuf, value *val, int size, bool newline)
+{
+  if (size == 2 && port->use_dw_for_init)
+    {
+      dbuf_tprintf (oBuf, "\t!dws\n", aopLiteralLong (val, 0, 2));
+      return;
+    }
+
+  const bool use_ret = TARGET_PDK_LIKE && !TARGET_IS_PDK16;
+
+  if (!use_ret)
+    dbuf_printf (oBuf, "\t.byte ");
+
+  for (int i = 0; i < size; i++)
+    {
+      const char *inst;
+      const char *byte = aopLiteral (val, port->little_endian ? i : size - 1 - i);
+      if (use_ret)
+        inst = i != size - 1 ? "\tret %s\n" : "\tret %s";
+      else
+        inst = i != size - 1 ? "%s, " : "%s";
+      dbuf_printf (oBuf, inst, byte);
+    }
+  if (newline)
+    dbuf_printf (oBuf, "\n");
+}
 
 /*-----------------------------------------------------------------*/
 /* _printPointerType - generates ival for pointer type             */
@@ -752,6 +782,8 @@ pointerTypeToGPByte (const int p_type, const char *iname, const char *oname)
 static void
 _printPointerType (struct dbuf_s *oBuf, const char *name, int size)
 {
+  wassert (!TARGET_PDK_LIKE);
+
   if (size == 4)
     {
       if (port->little_endian)
@@ -897,19 +929,10 @@ printIvalType (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s *o
     case 2:
       if (port->use_dw_for_init)
         {
-          dbuf_tprintf (oBuf, "\t!dws\n", aopLiteralLong (val, 0, 2));
+          printIvalVal (oBuf, val, 2, true);
           break;
         }
-      else if (port->little_endian)
-        {
-          dbuf_printf (oBuf, "\t.byte %s,%s",
-                       aopLiteral (val, 0), aopLiteral (val, 1));
-        }
-      else
-        {
-          dbuf_printf (oBuf, "\t.byte %s,%s",
-                       aopLiteral (val, 1), aopLiteral (val, 0));
-        }
+      printIvalVal (oBuf, val, 2, false);
       if (IS_UNSIGNED (val->type))
         dbuf_printf (oBuf, "\t; %u\n", (unsigned int) ulVal);
       else
@@ -924,16 +947,8 @@ printIvalType (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s *o
         }
       else
         {
-          if (port->little_endian)
-            {
-              dbuf_printf (oBuf, "\t.byte %s,%s,%s,%s",
-                           aopLiteral (val, 0), aopLiteral (val, 1), aopLiteral (val, 2), aopLiteral (val, 3));
-            }
-          else
-            {
-              dbuf_printf (oBuf, "\t.byte %s,%s,%s,%s",
-                           aopLiteral (val, 3), aopLiteral (val, 2), aopLiteral (val, 1), aopLiteral (val, 0));
-            }
+          printIvalVal (oBuf, val, 4, false);
+
           if (IS_FLOAT (val->type))
             {
               dbuf_printf (oBuf, "\t; % e\n", floatFromVal (val));
@@ -948,18 +963,7 @@ printIvalType (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s *o
         }
       break;
     case 8:
-      if (port->little_endian)
-        {
-          dbuf_printf (oBuf, "\t.byte %s,%s,%s,%s,%s,%s,%s,%s",
-            aopLiteral (val, 0), aopLiteral (val, 1), aopLiteral (val, 2), aopLiteral (val, 3), aopLiteral (val, 4), aopLiteral (val, 5), aopLiteral (val, 6), aopLiteral (val, 7));
-        }
-      else
-        {
-          dbuf_printf (oBuf, "\t.byte %s,%s,%s,%s,%s,%s,%s,%s",
-            aopLiteral (val, 7), aopLiteral (val, 6), aopLiteral (val, 5), aopLiteral (val, 4), aopLiteral (val, 3), aopLiteral (val, 2), aopLiteral (val, 1), aopLiteral (val, 0));
-        }
-      // TODO: Print value as comment. Does dbuf_printf support long long even on MSVC?
-      dbuf_printf (oBuf, "\n");
+      printIvalVal (oBuf, val, 8, true); // TODO: Print value as comment. Does dbuf_printf support long long even on MSVC?
       break;
     default:
       wassertl (0, "Attempting to initialize integer of non-handled size.");
@@ -968,8 +972,9 @@ printIvalType (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s *o
 
 /*-----------------------------------------------------------------*/
 /* printIvalBitFields - generate initializer for bitfields         */
+/* return the number of bytes written                              */
 /*-----------------------------------------------------------------*/
-static void
+static unsigned int
 printIvalBitFields (symbol ** sym, initList ** ilist, struct dbuf_s *oBuf)
 {
   symbol *lsym = *sym;
@@ -977,6 +982,7 @@ printIvalBitFields (symbol ** sym, initList ** ilist, struct dbuf_s *oBuf)
   unsigned long ival = 0;
   unsigned size = 0;
   unsigned bit_start = 0;
+  unsigned long int bytes_written = 0;
 
   while (lsym && IS_BITFIELD (lsym->type))
     {
@@ -1025,31 +1031,49 @@ printIvalBitFields (symbol ** sym, initList ** ilist, struct dbuf_s *oBuf)
 
   switch (size)
     {
+    case 0:
+      dbuf_tprintf (oBuf, "\n");
+      break;
+
     case 1:
       dbuf_tprintf (oBuf, "\t!db !constbyte\n", ival);
+      bytes_written++;
       break;
 
     case 2:
-      dbuf_tprintf (oBuf, "\t!db !constbyte, !constbyte\n", (ival & 0xff), (ival >> 8) & 0xff);
+      if (TARGET_PDK_LIKE && !TARGET_IS_PDK16)
+        {
+          dbuf_tprintf (oBuf, "\t!db !constbyte\n", (ival & 0xff));
+          dbuf_tprintf (oBuf, "\t!db !constbyte\n", ((ival >> 8) & 0xff));
+        }
+      else
+        dbuf_tprintf (oBuf, "\t!db !constbyte, !constbyte\n", (ival & 0xff), (ival >> 8) & 0xff);
+      bytes_written += 2;
       break;
 
     case 4:
       dbuf_tprintf (oBuf, "\t!db !constbyte, !constbyte, !constbyte, !constbyte\n",
                     (ival & 0xff), (ival >> 8) & 0xff, (ival >> 16) & 0xff, (ival >> 24) & 0xff);
+      bytes_written += 4;
       break;
+    default:
+      wassert (0);
     }
   *sym = lsym;
   *ilist = lilist;
+
+  return (bytes_written);
 }
 
 /*-----------------------------------------------------------------*/
 /* printIvalStruct - generates initial value for structures        */
 /*-----------------------------------------------------------------*/
 static void
-printIvalStruct (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s *oBuf)
+printIvalStruct (symbol *sym, sym_link *type, initList *ilist, struct dbuf_s *oBuf)
 {
   symbol *sflds;
   initList *iloop = NULL;
+  unsigned int skip_holes = 0;
 
   sflds = SPEC_STRUCT (type)->fields;
 
@@ -1092,6 +1116,8 @@ printIvalStruct (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s
     {
       while (sflds)
         {
+          unsigned int oldoffset = sflds->offset;
+
           if (IS_BITFIELD (sflds->type))
             printIvalBitFields (&sflds, &iloop, oBuf);
           else
@@ -1100,6 +1126,19 @@ printIvalStruct (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s
               sflds = sflds->next;
               iloop = iloop ? iloop->next : NULL;
             }
+
+          // Handle members from anonymous unions. Just a hack to fix bug #2643.
+          while (sflds && sflds->offset == oldoffset)
+            {
+              sflds = sflds->next;
+              skip_holes++;
+            }
+        }
+
+      while (skip_holes && iloop && iloop->type == INIT_HOLE)
+        {
+          skip_holes--;
+          iloop = iloop ? iloop->next : NULL;
         }
     }
 
@@ -1241,7 +1280,7 @@ printIvalChar32 (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s
     {
       val = list2val (ilist, TRUE);
       /* if the value is a character string  */
-      if (IS_ARRAY (val->type) && IS_INT (val->etype) && IS_UNSIGNED (val->etype) && !IS_LONG (val->etype))
+      if (IS_ARRAY (val->type) && IS_INT (val->etype) && IS_UNSIGNED (val->etype) && IS_LONG (val->etype))
         {
           if (!size)
             {
@@ -1427,9 +1466,19 @@ printIvalFuncPtr (sym_link * type, initList * ilist, struct dbuf_s *oBuf)
 
   size = getSize (type);
 
-  if (size == FPTRSIZE)
+  if (size == FUNCPTRSIZE)
     {
-      if (port->use_dw_for_init)
+      if (TARGET_PDK_LIKE)
+        {
+          dbuf_printf (oBuf, "\tret #<%s\n", name);
+          dbuf_printf (oBuf, "\tret #>%s\n", name);
+        }
+      else if (TARGET_IS_STM8 && FUNCPTRSIZE == 3)
+        {
+          _printPointerType (oBuf, name, size);
+          dbuf_printf (oBuf, "\n");
+        }
+      else if (port->use_dw_for_init)
         {
           dbuf_tprintf (oBuf, "\t!dws\n", name);
         }
@@ -1438,14 +1487,14 @@ printIvalFuncPtr (sym_link * type, initList * ilist, struct dbuf_s *oBuf)
           printPointerType (oBuf, name);
         }
     }
-  else if (size == GPTRSIZE)
+  else if (size == BFUNCPTRSIZE)
     {
       _printPointerType (oBuf, name, size);
       dbuf_printf (oBuf, "\n");
     }
   else
     {
-      assert (0);
+      wassertl (0, "Invalid function pointer size.");
     }
 
   return;
@@ -1478,16 +1527,17 @@ printIvalCharPtr (symbol * sym, sym_link * type, value * val, struct dbuf_s *oBu
         {
           dbuf_tprintf (oBuf, "\t!dbs\n", val->name);
         }
-      else if (size == FPTRSIZE)
+      else if (size == FARPTRSIZE)
         {
-          if (port->use_dw_for_init)
+          if (TARGET_PDK_LIKE && !TARGET_IS_PDK16)
             {
-              dbuf_tprintf (oBuf, "\t!dws\n", val->name);
+              dbuf_printf (oBuf, "\tret #<%s\n", val->name);
+              dbuf_printf (oBuf, IN_CODESPACE (SPEC_OCLS (val->etype)) ? "\tret #>(%s + 0x8000)\n" : "\tret #0\n", val->name);
             }
+          else if (port->use_dw_for_init)
+            dbuf_tprintf (oBuf, "\t!dws\n", val->name);
           else
-            {
-              printPointerType (oBuf, val->name);
-            }
+            printPointerType (oBuf, val->name);
         }
       else if (size == GPTRSIZE)
         {
@@ -1521,7 +1571,12 @@ printIvalCharPtr (symbol * sym, sym_link * type, value * val, struct dbuf_s *oBu
           dbuf_tprintf (oBuf, "\t!dbs\n", aopLiteral (val, 0));
           break;
         case 2:
-          if (port->use_dw_for_init)
+          if (TARGET_PDK_LIKE && !TARGET_IS_PDK16)
+            {
+              dbuf_tprintf (oBuf, "\tret %s\n", aopLiteral (val, 0));
+              dbuf_tprintf (oBuf, "\tret %s\n", aopLiteral (val, 1));
+            }
+          else if (port->use_dw_for_init)
             dbuf_tprintf (oBuf, "\t!dws\n", aopLiteralLong (val, 0, size));
           else if (port->little_endian)
             dbuf_tprintf (oBuf, "\t.byte %s,%s\n", aopLiteral (val, 0), aopLiteral (val, 1));
@@ -1529,7 +1584,7 @@ printIvalCharPtr (symbol * sym, sym_link * type, value * val, struct dbuf_s *oBu
             dbuf_tprintf (oBuf, "\t.byte %s,%s\n", aopLiteral (val, 1), aopLiteral (val, 0));
           break;
         case 3:
-          if (IS_GENPTR (type) && GPTRSIZE > FPTRSIZE && floatFromVal (val) != 0)
+          if (IS_GENPTR (type) && GPTRSIZE > FARPTRSIZE && floatFromVal (val) != 0)
             {
               if (!IS_PTR (val->type) && !IS_FUNC (val->type))
                 {
@@ -1550,7 +1605,7 @@ printIvalCharPtr (symbol * sym, sym_link * type, value * val, struct dbuf_s *oBu
             }
           break;
         case 4:
-          if (IS_GENPTR (type) && GPTRSIZE > FPTRSIZE && floatFromVal (val) != 0)
+          if (IS_GENPTR (type) && GPTRSIZE > FARPTRSIZE && floatFromVal (val) != 0)
             {
               if (!IS_PTR (val->type) && !IS_FUNC (val->type))
                 {
@@ -1605,7 +1660,7 @@ printIvalCharPtr (symbol * sym, sym_link * type, value * val, struct dbuf_s *oBu
 /* printIvalPtr - generates initial value for pointers             */
 /*-----------------------------------------------------------------*/
 void
-printIvalPtr (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s *oBuf)
+printIvalPtr (symbol *sym, sym_link *type, initList *ilist, struct dbuf_s *oBuf)
 {
   value *val;
   int size;
@@ -1681,11 +1736,24 @@ printIvalPtr (symbol * sym, sym_link * type, initList * ilist, struct dbuf_s *oB
 
   size = getSize (type);
 
-  if (size == 1)                /* Z80 specific?? */
+  if (TARGET_PDK_LIKE && size == 2)
+    {
+      if (IN_CODESPACE (SPEC_OCLS (val->etype)))
+        {
+          dbuf_printf (oBuf, "\tret #<%s\n", val->name);
+          dbuf_printf (oBuf, "\tret #>(%s + 0x8000)\n", val->name);
+        }
+      else
+        {
+          dbuf_printf (oBuf, "\tret #%s\n", val->name);
+          dbuf_printf (oBuf, "\tret #0\n");
+        }
+    }
+  else if (size == 1)                /* Z80 specific?? */
     {
       dbuf_tprintf (oBuf, "\t!dbs\n", val->name);
     }
-  else if (size == FPTRSIZE)
+  else if (size == FARPTRSIZE)
     {
       if (port->use_dw_for_init)
         dbuf_tprintf (oBuf, "\t!dws\n", val->name);
@@ -1893,13 +1961,15 @@ emitStaticSeg (memmap *map, struct dbuf_s *oBuf)
                   emitDebugSym (oBuf, sym);
                   dbuf_printf (oBuf, " == .\n");
                 }
-              dbuf_printf (oBuf, "%s:\n", sym->rname);
               /* special case for character strings */
               if (IS_ARRAY (sym->type) &&
                 (IS_CHAR (sym->type->next) && SPEC_CVAL (sym->etype).v_char ||
                  IS_INT (sym->type->next) && !IS_LONG (sym->type->next) && SPEC_CVAL (sym->etype).v_char16 ||
                  IS_INT (sym->type->next) && IS_LONG (sym->type->next) && SPEC_CVAL (sym->etype).v_char32))
                 {
+                  if (options.const_seg)
+                    dbuf_tprintf(&code->oBuf, "\t!area\n", options.const_seg);
+                  dbuf_printf (oBuf, "%s:\n", sym->rname);
                   if (IS_CHAR (sym->type->next))
                     printChar (oBuf, SPEC_CVAL (sym->etype).v_char, size);
                   else if (IS_INT (sym->type->next) && !IS_LONG (sym->type->next))
@@ -1908,9 +1978,12 @@ emitStaticSeg (memmap *map, struct dbuf_s *oBuf)
                     printChar32 (oBuf, SPEC_CVAL (sym->etype).v_char32, size / 4);
                   else
                     wassert (0);
+                  if (options.const_seg)
+                    dbuf_tprintf(oBuf, "\t!areacode\n", options.code_seg);
                 }
               else
                 {
+                  dbuf_printf (oBuf, "%s:\n", sym->rname);
                   dbuf_tprintf (oBuf, "\t!ds\n", (unsigned int) size & 0xffff);
                 }
             }
@@ -1936,8 +2009,8 @@ void
 emitMaps (void)
 {
   namedspacemap *nm;
-  int publicsfr = TARGET_IS_MCS51;      /* Ideally, this should be true for all  */
-                                        /* ports but let's be conservative - EEP */
+  int publicsfr = TARGET_IS_MCS51 || TARGET_PDK_LIKE;      /* Ideally, this should be true for all  */
+                                                           /* ports but let's be conservative - EEP */
 
   inInitMode++;
   /* no special considerations for the following
@@ -1969,10 +2042,9 @@ emitMaps (void)
   emitRegularMap (code, TRUE, FALSE);
 
   if (options.const_seg)
-    {
-      dbuf_tprintf (&code->oBuf, "\t!area\n", options.const_seg);
-    }
+    dbuf_tprintf (&code->oBuf, "\t!area\n", options.const_seg);
   emitStaticSeg (statsg, &code->oBuf);
+
   if (port->genXINIT)
     {
       dbuf_tprintf (&code->oBuf, "\t!area\n", xinit->sname);
@@ -2184,7 +2256,7 @@ glue (void)
   dbuf_init (&ovrBuf, 4096);
 
   mcs51_like = (port->general.glue_up_main &&
-                (TARGET_IS_MCS51 || TARGET_IS_DS390 || TARGET_IS_XA51 || TARGET_IS_DS400));
+                (TARGET_IS_MCS51 || TARGET_IS_DS390 || TARGET_IS_DS400));
 
   /* print the global struct definitions */
   if (options.debug)
@@ -2236,6 +2308,8 @@ glue (void)
     fprintf (asmFile, "\t.hd64\n");
   else if (TARGET_IS_R3KA)
     fprintf (asmFile, "\t.r3k\n");
+  else if (TARGET_IS_EZ80_Z80)
+    fprintf (asmFile, "\t.ez80\n");
 
   /* print module name */
   tfprintf (asmFile, "\t!module\n", moduleName);
@@ -2298,7 +2372,7 @@ glue (void)
   if (port->assembler.externGlobal)
     printExterns (asmFile);
 
-  if ((mcs51_like) || (TARGET_IS_Z80 || TARGET_IS_Z180 || TARGET_IS_RABBIT))  /*.p.t.20030924 need to output SFR table for Z80 as well */
+  if ((mcs51_like) || (TARGET_IS_Z80 || TARGET_IS_GBZ80 || TARGET_IS_Z180 || TARGET_IS_RABBIT || TARGET_IS_EZ80_Z80) || TARGET_PDK_LIKE)  /*.p.t.20030924 need to output SFR table for Z80 as well */
     {
       /* copy the sfr segment */
       fprintf (asmFile, "%s", iComments2);
@@ -2509,7 +2583,7 @@ glue (void)
     }
   dbuf_write_and_destroy (&statsg->oBuf, asmFile);
 
-  /* STM8 note: there are no such instructions supported.
+  /* STM8 / PDK14 note: there are no such instructions supported.
      Also, we don't need this logic as well. */
   if (port->general.glue_up_main && mainf && IFFUNC_HASBODY (mainf->type))
     {
@@ -2520,6 +2594,8 @@ glue (void)
       tfprintf (asmFile, "\t!area\n", port->mem.post_static_name);
       if(TARGET_IS_STM8)
         fprintf (asmFile, "\tjp\t__sdcc_program_startup\n");
+      else if(TARGET_PDK_LIKE)
+        fprintf (asmFile, "\tgoto\t__sdcc_program_startup\n");
       else
         fprintf (asmFile, "\t%cjmp\t__sdcc_program_startup\n", options.acall_ajmp ? 'a' : 'l');
     }
@@ -2540,7 +2616,9 @@ glue (void)
 
       /* put in jump or call to main */
       if(TARGET_IS_STM8)
-        fprintf (asmFile, "\tjp\t_main\n");
+        fprintf (asmFile, options.model == MODEL_LARGE ? "\tjpf\t_main\n" : "\tjp\t_main\n");
+      else if(TARGET_PDK_LIKE)
+        fprintf (asmFile, "\tgoto\t_main\n");
       else
         fprintf (asmFile, "\t%cjmp\t_main\n", options.acall_ajmp ? 'a' : 'l');        /* needed? */
       fprintf (asmFile, ";\treturn from main will return to caller\n");
