@@ -31,7 +31,7 @@
 #include "newalloc.h"
 #include "dbuf_string.h"
 
-int cNestLevel;
+long cNestLevel;
 
 
 /*-----------------------------------------------------------------*/
@@ -299,9 +299,9 @@ list2expr (initList * ilist)
 /* resolveIvalSym - resolve symbols in initial values               */
 /*------------------------------------------------------------------*/
 void
-resolveIvalSym (initList * ilist, sym_link * type)
+resolveIvalSym (initList *ilist, sym_link *type)
 {
-  int is_ptr = IS_PTR (type);
+  int is_ptr = IS_PTR (type) || (IS_ARRAY(type) && IS_PTR(type->next));
   RESULT_TYPE resultType = getResultTypeFromType (getSpec (type));
 
   while (ilist)
@@ -1120,27 +1120,63 @@ constFixed16x16Val (const char *s)
 }
 
 /*-----------------------------------------------------------------*/
-/* constVal - converts an INTEGER constant into a cheapest value   */
+/* constVal - converts a constant into a cheap value type          */
 /*-----------------------------------------------------------------*/
 value *
 constVal (const char *s)
 {
-  value *val;
+  value *val = constIntVal (s);
+
+  wassert (SPEC_NOUN (val->type) == V_INT);
+
+  if (SPEC_LONGLONG (val->type))
+    ;
+  else if (SPEC_LONG (val->type))
+    ;
+  else if (SPEC_USIGN (val->type))
+    {
+      unsigned int i = SPEC_CVAL (val->type).v_uint;
+      if (i < 256)
+        SPEC_NOUN (val->type) = V_CHAR;
+    }
+  else
+    {
+      int i = SPEC_CVAL (val->type).v_int;
+      if (i >= 0 && i < 256)
+        {
+          SPEC_NOUN (val->type) = V_CHAR;
+          SPEC_USIGN (val->type) = TRUE;
+          SPEC_CVAL (val->type).v_uint = i;
+        }
+      else if (i >= -128 && i < 128)
+        {
+          SPEC_NOUN (val->type) = V_CHAR;
+        }
+    }
+
+  return val;
+}
+
+/*-----------------------------------------------------------------*/
+/* constIntVal - converts an integer constant into correct type    */
+/* See ISO C11, section 6.4.4.1 for the rules.                     */
+/*-----------------------------------------------------------------*/
+value *
+constIntVal (const char *s)
+{
   char *p, *p2;
   double dval;
   long long int llval;
-  bool is_integral = 0;
+  value *val = newValue ();
+  bool decimal, u_suffix = FALSE, l_suffix = FALSE, ll_suffix = FALSE;
 
-  val = newValue ();            /* alloc space for value */
-
-  val->type = val->etype = newLink (SPECIFIER); /* create the specifier */
+  val->type = val->etype = newLink (SPECIFIER);
   SPEC_SCLS (val->type) = S_LITERAL;
   SPEC_CONST (val->type) = 1;
-  /* let's start with a signed char */
-  SPEC_NOUN (val->type) = V_CHAR;
   SPEC_USIGN (val->type) = 0;
 
   errno = 0;
+
   if (s[0] == '0')
     {
       if (s[1] == 'b' || s[1] == 'B')
@@ -1148,7 +1184,7 @@ constVal (const char *s)
       else
         llval = strtoull (s, &p, 0);
       dval = (double)(unsigned long long int) llval;
-      is_integral = 1;
+      decimal = FALSE;
     }
   else
     {
@@ -1157,6 +1193,7 @@ constVal (const char *s)
         llval = strtoull (s, &p, 0);
       else
         llval = strtoll (s, &p, 0);
+      decimal = TRUE;
     }
 
   if (errno)
@@ -1165,11 +1202,10 @@ constVal (const char *s)
       werror (W_INVALID_INT_CONST, s, dval);
     }
 
-  /* Setup the flags first */
-  /* set the unsigned flag if 'uU' is found */
+  // Check suffixes
   if ((p2 = strchr (p, 'u')) || (p2 = strchr (p, 'U')))
     {
-      SPEC_USIGN (val->type) = 1;
+      u_suffix = TRUE;
       p2++;
       if (strchr (p2, 'u') || strchr (p2, 'U'))
         werror (E_INTEGERSUFFIX, p);
@@ -1177,89 +1213,71 @@ constVal (const char *s)
 
   if ((p2 = strstr (p, "ll")) || (p2 = strstr (p, "LL")))
     {
-      SPEC_NOUN (val->type) = V_INT;
-      SPEC_LONGLONG (val->type) = 1;
+      ll_suffix = TRUE;
       p2 += 2;
       if (strchr (p2, 'l') || strchr (p2, 'L'))
-        werror (E_INTEGERSUFFIX, p); 
+        werror (E_INTEGERSUFFIX, p);
     }
-  /* set the b_long flag if 'l' or 'L' is found */
   else if ((p2 = strchr (p, 'l')) || (p2 = strchr (p, 'L')))
     {
-      SPEC_NOUN (val->type) = V_INT;
-      SPEC_LONG (val->type) = 1;
+      l_suffix = TRUE;
       p2++;
       if (strchr (p2, 'l') || strchr (p2, 'L'))
         werror (E_INTEGERSUFFIX, p);
-      /* c89 allows unsigned long without explicit u suffix */
-      if (!options.std_c99 && dval > 0x7fffffff && dval <= 0xffffffff)
-         SPEC_USIGN (val->type) = 1;
+    }
+
+  SPEC_NOUN (val->type) = V_INT;
+
+  if (u_suffix) // Choose first of unsigned int, unsigned long int, unsigned long long int that fits.
+    {
+      SPEC_USIGN (val->type) = 1;
+      if (ll_suffix || dval > 0xffffffff)
+        SPEC_LONGLONG (val->type) = 1;
+      else if(l_suffix || dval > 0xffff)
+        SPEC_LONG (val->type) = 1;
     }
   else
-    {
-      if (dval < 0)
-        {                       /* "-28u" will still be signed and negative */
-          if (dval < -128)
-            {                   /* check if we have to promote to int */
-              SPEC_NOUN (val->type) = V_INT;
-            }
-          if (dval < -32768)
-            {                   /* check if we have to promote to long int */              
-              SPEC_LONG (val->type) = 1;
-            }
-          if (dval < -2147483648.0)
-            {                   /* check if we have to promote to long long int */
-              SPEC_LONGLONG (val->type) = 1;
-            }
-        }
-      else
-        {                       /* >=0 */
-          if (dval > 0xff ||    /* check if we have to promote to int */
-              SPEC_USIGN (val->type))
-            {                   /* If it's unsigned, we can't use unsigned char.
-                                   After an integral promotion it will be a signed int;
-                                   this certainly isn't what the programer wants */
-              SPEC_NOUN (val->type) = V_INT;
-            }
-          else
-            {                   /* store char's always as unsigned; this helps other optimizations */
-              SPEC_USIGN (val->type) = 1;
-            }
-          if (dval > 0xffff && SPEC_USIGN (val->type) && !SPEC_LONGLONG (val->type))
-            {                   /* check if we have to promote to long */
-              SPEC_LONG (val->type) = 1;
-            }
-          else if (dval > 0x7fff && !SPEC_USIGN (val->type))
-            {                   /* check if we have to promote to long int */
-              if (is_integral && dval <= 0xffff)
-                {               /* integral (hex, octal and binary) constants may be stored in unsigned type */
-                  SPEC_USIGN (val->type) = 1;
-                }
-              else
+    {  
+      if (decimal) // Choose first of int, long int, long long int that fits.
+        {
+          if (ll_suffix || dval > 0x7fffffff || dval < -0x80000000ll)
+            {
+              if (!options.std_c99) // C90 exception: Use unsigned long
                 {
+                  SPEC_USIGN (val->type) = 1;
                   SPEC_LONG (val->type) = 1;
                 }
-            }
-          if (dval > 0xffffffff && SPEC_USIGN (val->type) && !SPEC_LONGLONG (val->type))
-            {
-              SPEC_LONGLONG (val->type) = 1;
-            }
-          else if (dval > 0x7fffffff && !SPEC_USIGN (val->type))
-            {
-              /* integral constants can be unsigned long. */
-              /* c89 also allows unsigned long decimal constants without explicit suffix */
-              if ((is_integral || !options.std_c99) && dval <= 0xffffffff)
-                {
-                  SPEC_USIGN (val->type) = 1;
-                }
               else
-                {
-                  SPEC_LONGLONG (val->type) = 1;
-                  if (is_integral && (llval & 0x8000000000000000ull))
-                    SPEC_USIGN (val->type) = 1;
-                }
+                SPEC_LONGLONG (val->type) = 1;
             }
+          else if(l_suffix || dval > 0x7fff || dval < -0x8000l)
+            SPEC_LONG (val->type) = 1;
         }
+      else // Choose first of int, unsigned int, long int, unsigned long int, long long int, unsigned long long int that fits.
+       {
+         if (dval > 0x7fffffffffffffff)
+           {
+             SPEC_USIGN (val->type) = 1;
+             SPEC_LONGLONG (val->type) = 1;
+           }
+         else if (ll_suffix || dval > 0xffffffff || dval < -0x80000000ll)
+           {
+             SPEC_LONGLONG (val->type) = 1;
+           }
+         else if (dval > 0x7fffffff)
+           {
+             SPEC_USIGN (val->type) = 1;
+             SPEC_LONG (val->type) = 1;
+           }
+         else if (l_suffix || dval > 0xffff || dval < -0x8000l)
+           {
+             SPEC_LONG (val->type) = 1;
+           }
+         else if (dval > 0x7fff)
+           {
+             SPEC_USIGN (val->type) = 1;
+           }
+       }
     }
 
   /* check for out of range */
@@ -1487,7 +1505,7 @@ strVal (const char *s)
   const char *utf_8;
   size_t utf_8_size;
 
-  val = newValue ();            /* get a new one */
+  val = newValue ();
 
   /* get a declarator */
   val->type = newLink (DECLARATOR);
@@ -1502,6 +1520,8 @@ strVal (const char *s)
   if (s[0] == '"') // UTF-8 string literal (any prefix u8 or L in the source would already have been stripped by earlier stages)
     {
       SPEC_NOUN (val->etype) = V_CHAR;
+      SPEC_USIGN (val->etype) = !options.signed_char;
+      val->etype->select.s.b_implicit_sign = true;
       SPEC_CVAL (val->etype).v_char = utf_8;
       DCL_ELEM (val->type) = utf_8_size;
     }
@@ -1535,7 +1555,35 @@ strVal (const char *s)
         wassert (0);
     }
 
-  return val;
+  return (val);
+}
+
+/*------------------------------------------------------------------*/
+/* rawStrVal - converts a string to a value                         */
+/*------------------------------------------------------------------*/
+value *
+rawStrVal (const char *s, size_t size)
+{
+  struct dbuf_s dbuf;
+  value *val = newValue ();
+  
+  dbuf_init (&dbuf, size);
+  wassert (dbuf_append (&dbuf, s, size));
+
+  /* get a declarator */
+  val->type = newLink (DECLARATOR);
+  DCL_TYPE (val->type) = ARRAY;
+  val->type->next = val->etype = newLink (SPECIFIER);
+  SPEC_SCLS (val->etype) = S_LITERAL;
+  SPEC_CONST (val->etype) = 1;
+
+  SPEC_NOUN (val->etype) = V_CHAR;
+  SPEC_USIGN (val->etype) = !options.signed_char;
+  val->etype->select.s.b_implicit_sign = true;
+  SPEC_CVAL (val->etype).v_char = dbuf_detach (&dbuf);
+  DCL_ELEM (val->type) = size;
+
+  return (val);
 }
 
 /*------------------------------------------------------------------*/
@@ -1690,8 +1738,16 @@ charVal (const char *s)
           return constCharacterVal (*s, type);
         }
     }
-  else                          /* not a backslash */
-    return constCharacterVal (*s, type);
+  else if (type) // Wide character constant
+    {
+      size_t ulen;
+      const TYPE_UDWORD *ustr = utf_32_from_utf_8 (&ulen, s, strlen(s) - 1);
+      value *val = constCharacterVal (*ustr, type);
+      free ((void *)ustr);
+      return (val);
+    }
+  else // Character constant that is not wide - compability with legacy encodings.
+    return constCharacterVal (*s, 0);
 }
 
 /*------------------------------------------------------------------*/
@@ -1782,7 +1838,7 @@ floatFromVal (value * val)
 /* ulFromVal - value to unsigned long conversion                    */
 /*------------------------------------------------------------------*/
 unsigned long
-ulFromVal (value * val)
+ulFromVal (value *val)
 {
   if (!val)
     return 0;
@@ -1862,6 +1918,8 @@ byteOfVal (value * val, int offset)
   unsigned char *p;
   int shift = 8*offset;
 
+  wassert (offset >= 0);
+
   if (!val)
     return 0;
 
@@ -2030,6 +2088,44 @@ ullFromVal (value * val)
   return (unsigned long long) ullFromLit (val->type);
 }
 
+/*------------------------------------------------------------------*/
+/* csdOfVal - return 0 if the value can be represented as csd       */
+/* topbit  - highest nonzero bit in csd                             */
+/* nonzero - number of nonzero bits in csd                          */
+/* csd_add - positive bits in csd                                   */
+/* csd_sub - negative bits in csd                                   */
+/*------------------------------------------------------------------*/
+int csdOfVal (int *topbit, int *nonzero, unsigned long long *csd_add, unsigned long long *csd_sub, value *val)
+{
+  unsigned long long binary = ullFromVal (val);
+  bool gamma, theta, a;
+  int bit, next;
+
+  *topbit = 0;
+  *nonzero = 0;
+  *csd_add = 0;
+  *csd_sub = 0;
+
+  for (a = 0, gamma = 0, bit = 0; bit < 61; bit++)
+    {
+       theta = a ^ (binary & 1);
+       gamma = !gamma && theta;
+       next = (1 - 2 * (bool)(binary & 2)) * gamma;
+       if (next > 0)
+         *csd_add |= (1ull << bit);
+       else if (next < 0)
+         *csd_sub |= (1ull << bit);
+       if (next)
+         {
+           (*nonzero)++;
+           *topbit = bit;
+         }
+       a = (binary & 1);
+       binary >>= 1;
+    }
+  return((bool)binary);
+}
+
 /*------------------------------------------------------------------*/
 /* isEqualVal - return 1 if value is equal to specified constant    */
 /*------------------------------------------------------------------*/
@@ -2271,7 +2367,7 @@ valDiv (value * lval, value * rval)
 {
   value *val;
 
-  if (isEqualVal (rval, 0))
+  if (isEqualVal (rval, 0) && !IS_FLOAT (computeType (lval->etype, rval->etype, RESULT_TYPE_INT, '/')))
     {
       werror (E_DIVIDE_BY_ZERO);
       return rval;
@@ -3178,8 +3274,8 @@ valForCastAggr (ast * aexpr, sym_link * type, ast * cnst, int op)
 }
 
 /*-----------------------------------------------------------------*/
-/* valForCastAggr - will return value for a cast of an aggregate   */
-/*                  with no constant                               */
+/* valForCastArr - will return value for a cast of an aggregate    */
+/*                 with no constant                                */
 /*-----------------------------------------------------------------*/
 value *
 valForCastArr (ast * aexpr, sym_link * type)
