@@ -7,13 +7,15 @@ extern void genSTM08Code (iCode *);
 
 reg_info stm8_regs[] =
 {
-
   {REG_GPR, A_IDX,   "a"},
   {REG_GPR, XL_IDX,  "xl"},
   {REG_GPR, XH_IDX,  "xh"},
   {REG_GPR, YL_IDX,  "yl"},
   {REG_GPR, YH_IDX,  "yh"},
   {REG_CND, C_IDX,   "c"},
+  {REG_GPR, X_IDX,   "x"},
+  {REG_GPR, Y_IDX,   "y"},
+  {0, SP_IDX,        "sp"},
 };
 
 /* Flags to turn on debugging code.
@@ -97,24 +99,7 @@ createStackSpil (symbol * sym)
   symbol *sloc = NULL;
   struct dbuf_s dbuf;
 
-  D (D_ALLOC, ("createStackSpil: for sym %p %s\n", sym, sym->name));
-
-  /* first go try and find a free one that is already
-     existing on the stack */
-  if (applyToSet (_G.stackSpil, isFreeSTM8, &sloc, sym))
-    {
-      /* found a free one : just update & return */
-      sym->usl.spillLoc = sloc;
-      sym->stackSpil = 1;
-      sloc->isFree = 0;
-      addSetHead (&sloc->usl.itmpStack, sym);
-      D (D_ALLOC, ("createStackSpil: found existing\n"));
-      return sym;
-    }
-
-  /* could not then have to create one , this is the hard part
-     we need to allocate this on the stack : this is really a
-     hack!! but cannot think of anything better at this time */
+  D (D_ALLOC, ("createStackSpil: for sym %p %s (old currFunc->stack %ld)\n", sym, sym->name, (long)(currFunc->stack)));
 
   dbuf_init (&dbuf, 128);
   dbuf_printf (&dbuf, "sloc%d", _G.slocNum++);
@@ -156,21 +141,19 @@ createStackSpil (symbol * sym)
      of the spill location */
   addSetHead (&sloc->usl.itmpStack, sym);
 
-  D (D_ALLOC, ("createStackSpil: created new\n"));
+  D (D_ALLOC, ("createStackSpil: created new %s\n", sloc->name));
   return sym;
 }
 
 /*-----------------------------------------------------------------*/
 /* spillThis - spils a specific operand                            */
 /*-----------------------------------------------------------------*/
-static void
-spillThis (symbol *sym, bool force_spill)
+void
+stm8SpillThis (symbol *sym, bool force_spill)
 {
   int i;
 
-  D (D_ALLOC, ("spillThis: spilling %p (%s)\n", sym, sym->name));
-
-  sym->for_newralloc = 0;
+  D (D_ALLOC, ("stm8SpillThis: spilling %p (%s)\n", sym, sym->name));
 
   /* if this is rematerializable or has a spillLocation
      we are okay, else we need to create a spillLocation
@@ -178,7 +161,7 @@ spillThis (symbol *sym, bool force_spill)
   if (!(sym->remat || sym->usl.spillLoc) || (sym->usl.spillLoc && !sym->usl.spillLoc->onStack)) // stm8 port currently only supports on-stack spill locations in code generation.
     createStackSpil (sym);
 
-  /* mark it has spilt & put it in the spilt set */
+  /* mark it as spilt */
   sym->isspilt = sym->spillA = 1;
 
   if (force_spill)
@@ -210,6 +193,12 @@ regTypeNum (void)
       /* if used zero times then no registers needed. Exception: Variables larger than 4 bytes - these might need a spill location when they are return values */
       if ((sym->liveTo - sym->liveFrom) == 0 && getSize (sym->type) <= 4)
         continue;
+      else if ((sym->liveTo - sym->liveFrom) == 0 && bitVectnBitsOn (sym->defs) <= 1)
+        {
+          iCode *dic = hTabItemWithKey (iCodehTab, bitVectFirstBit (sym->defs));
+          if (!dic || dic->op != CALL && dic->op != PCALL)
+            continue;
+        }
 
       D (D_ALLOC, ("regTypeNum: loop on sym %p\n", sym));
 
@@ -261,7 +250,28 @@ regTypeNum (void)
     }
 }
 
-#if 0
+/** Transform weird SDCC handling of writes via pointers
+    into something more sensible. */
+static void
+transformPointerSet (eBBlock **ebbs, int count)
+{
+  /* for all blocks */
+  for (int i = 0; i < count; i++)
+    {
+      iCode *ic;
+
+      /* for all instructions do */
+      for (ic = ebbs[i]->sch; ic; ic = ic->next)
+        if (POINTER_SET (ic))
+          {
+            IC_LEFT (ic) = IC_RESULT (ic);
+            IC_RESULT (ic) = 0;
+            ic->op = SET_VALUE_AT_ADDRESS;
+          }
+    }
+}
+
+
 /** Register reduction for assignment.
  */
 static int
@@ -310,9 +320,6 @@ packRegsForAssign (iCode *ic, eBBlock *ebp)
 
           if (IS_SYMOP (IC_RESULT (dic)) && IC_RESULT (dic)->key == IC_RIGHT (ic)->key)
             {
-              if (POINTER_SET (dic))
-                dic = NULL;
-
               break;
             }
 
@@ -342,7 +349,7 @@ packRegsForAssign (iCode *ic, eBBlock *ebp)
     return 0;                   /* did not find */
 
   /* if assignment then check that right is not a bit */
-  if (ASSIGNMENT (ic) && !POINTER_SET (ic))
+  if (ic->op == '=')
     {
       sym_link *etype = operandType (IC_RESULT (dic));
       if (IS_BITFIELD (etype))
@@ -356,8 +363,18 @@ packRegsForAssign (iCode *ic, eBBlock *ebp)
         }
     }
 
+  /* For now eliminate 8-bit temporary variables only.
+     The STM8 instructions operating directly on memory
+     operands are 8-bit, so the most benefit is in 8-bit
+     operations. On the other hand, supporting wider
+     operations well in codegen is also more effort. */
+  if (bitsForType (operandType (IC_RESULT (dic))) > 8 &&
+    !((dic->op == LEFT_OP || dic->op == RIGHT_OP) && IS_OP_LITERAL (IC_RIGHT (dic)) && operandLitValue (IC_RIGHT (dic))  == 1 && // Can do wide shift by 1 in place.
+      IS_SYMOP (IC_LEFT (dic)) && IS_SYMOP (IC_RESULT (ic)) && OP_SYMBOL (IC_LEFT (dic)) == OP_SYMBOL (IC_RESULT (ic))))
+    return 0;
+
   /* if the result is on stack or iaccess then it must be
-     the same atleast one of the operands */
+     the same as at least one of the operands */
   if (OP_SYMBOL (IC_RESULT (ic))->onStack || OP_SYMBOL (IC_RESULT (ic))->iaccess)
     {
       /* the operation has only one symbol
@@ -381,7 +398,7 @@ pack:
       OP_SYMBOL (IC_RESULT (dic))->liveFrom = dic->seq;
     }
   /* delete from liverange table also
-     delete from all the points inbetween and the new
+     delete from all the points in between and the new
      one */
   for (sic = dic; sic != ic; sic = sic->next)
     {
@@ -398,6 +415,84 @@ pack:
   return 1;
 }
 
+/** Will reduce some registers for single use.
+ */
+static int
+packRegsForOneuse (iCode *ic, operand **opp, eBBlock *ebp)
+{
+  iCode *dic;
+
+  operand *op = *opp;
+//printf("packRegsForOneuse() at ic %d\n", ic->key);
+  /* if returning a literal then do nothing */
+  if (!IS_ITEMP (op))
+    return 0;
+
+  /* if rematerializable do nothing */
+  if (OP_SYMBOL (op)->remat)
+    return 0;
+
+  /* this routine will mark the symbol as used in one
+     instruction use only && if the definition is local
+     (ie. within the basic block) && has only one definition  */
+  if (bitVectnBitsOn (OP_USES (op)) != 1 || bitVectnBitsOn (OP_DEFS (op)) != 1)
+    return 0;
+
+  /* get the definition */
+  if (!(dic = hTabItemWithKey (iCodehTab, bitVectFirstBit (OP_DEFS (op)))))
+    return 0;
+//printf("Found dic %d\n", dic->key);
+  /* found the definition now check if it is local */
+  if (dic->seq < ebp->fSeq || dic->seq > ebp->lSeq)
+    return 0;                /* non-local */
+
+  /* for now handle results from assignments from globals only */
+  if (dic->op != '=' || !isOperandGlobal (IC_RIGHT (dic)))
+    return 0;
+  /* also make sure the intervenening instructions
+     don't have any thing in far space */
+  for (iCode *nic = dic->next; nic && nic != ic; nic = nic->next)
+    {
+      /* if there is an intervening function call then no */
+      if (nic->op == CALL || nic->op == PCALL)
+        return 0;
+
+      if (nic->op == GET_VALUE_AT_ADDRESS || nic->op == SET_VALUE_AT_ADDRESS)
+        return 0;
+
+      /* if address of & the result is remat, then okay */
+      if (nic->op == ADDRESS_OF && OP_SYMBOL (IC_RESULT (nic))->remat)
+        continue;
+
+      if (IS_OP_VOLATILE (IC_LEFT (nic)) ||
+          IS_OP_VOLATILE (IC_RIGHT (nic)) ||
+          isOperandGlobal (IC_RESULT (nic)))
+        return 0;
+    }
+
+  /* Optimize out the assignment */
+  *opp = operandFromOperand (IC_RIGHT(dic));
+  (*opp)->isaddr = true;
+  
+  bitVectUnSetBit (OP_SYMBOL (op)->defs, dic->key);
+  bitVectUnSetBit (OP_SYMBOL (op)->uses, ic->key);
+
+  if (IS_ITEMP (IC_RESULT (dic)) && OP_SYMBOL (IC_RESULT (dic))->liveFrom > dic->seq)
+    OP_SYMBOL (IC_RESULT (dic))->liveFrom = dic->seq;
+
+  /* delete from liverange table also
+     delete from all the points in between and the new
+     one */
+  for (iCode *nic = dic; nic != ic; nic = nic->next) 
+    bitVectUnSetBit (nic->rlive, op->key);
+
+  remiCodeFromeBBlock (ebp, dic);
+
+  hTabDeleteItem (&iCodehTab, dic->key, ic, DELETE_ITEM, NULL);
+  
+  return 1;
+}
+
 /** Does some transformations to reduce register pressure.
  */
 static void
@@ -418,14 +513,88 @@ packRegisters (eBBlock * ebp)
       for (ic = ebp->sch; ic; ic = ic->next)
         {
           /* find assignment of the form TrueSym := iTempNN:1 */
-          if (ic->op == '=' && !POINTER_SET (ic))
+          if (ic->op == '=')
             change += packRegsForAssign (ic, ebp);
         }
       if (!change)
         break;
     }
+
+  for (ic = ebp->sch; ic; ic = ic->next)
+    {
+      D (D_ALLOC, ("packRegisters: looping on ic %p\n", ic));
+
+      /* Safe: address of a true sym is always constant. */
+      /* if this is an itemp & result of a address of a true sym
+         then mark this as rematerialisable   */
+      if (ic->op == ADDRESS_OF && 
+        IS_ITEMP (IC_RESULT (ic)) && bitVectnBitsOn (OP_DEFS (IC_RESULT (ic))) == 1 && !IS_PARM (IC_RESULT (ic)) /* The receiving of the parameter is not accounted for in DEFS */ &&
+        IS_TRUE_SYMOP (IC_LEFT (ic)) && !OP_SYMBOL (IC_LEFT (ic))->onStack)
+        {
+          OP_SYMBOL (IC_RESULT (ic))->remat = 1;
+          OP_SYMBOL (IC_RESULT (ic))->rematiCode = ic;
+          OP_SYMBOL (IC_RESULT (ic))->usl.spillLoc = NULL;
+        }
+
+      if (ic->op == ADDRESS_OF && IS_ITEMP (IC_RESULT (ic)) && bitVectnBitsOn (OP_DEFS (IC_RESULT (ic))) == 1 && !IS_PARM (IC_RESULT (ic)) && /* Can remat stack locations, but they can currently only be used for pointer read / write */
+        IS_TRUE_SYMOP (IC_LEFT (ic)) && OP_SYMBOL (IC_LEFT (ic))->onStack)
+        {
+          bool ok = true;
+          bitVect *uses = bitVectCopy (OP_USES (IC_RESULT (ic)));
+          for (int bit = bitVectFirstBit (uses); bitVectnBitsOn (uses); bitVectUnSetBit (uses, bit), bit = bitVectFirstBit (uses))
+            {
+              const iCode *uic = hTabItemWithKey (iCodehTab, bit);
+              wassert (uic);
+              if (uic->op != SET_VALUE_AT_ADDRESS && uic->op != GET_VALUE_AT_ADDRESS)
+                {
+                  ok = false;
+                  break;
+                }
+            }
+
+          if (ok)
+            {
+              OP_SYMBOL (IC_RESULT (ic))->remat = 1;
+              OP_SYMBOL (IC_RESULT (ic))->rematiCode = ic;
+              OP_SYMBOL (IC_RESULT (ic))->usl.spillLoc = NULL;
+            }
+
+          freeBitVect (uses);
+        }
+
+      /* Safe: just propagates the remat flag */
+      /* if straight assignment then carry remat flag if this is the
+         only definition */
+      if (ic->op == '=' && IS_SYMOP (IC_RIGHT (ic)) && OP_SYMBOL (IC_RIGHT (ic))->remat &&
+        !isOperandGlobal (IC_RESULT (ic)) && bitVectnBitsOn (OP_SYMBOL (IC_RESULT (ic))->defs) == 1 && !IS_PARM (IC_RESULT (ic)) /* The receiving of the paramter is not accounted for in DEFS */)
+        {
+          OP_SYMBOL (IC_RESULT (ic))->remat = OP_SYMBOL (IC_RIGHT (ic))->remat;
+          OP_SYMBOL (IC_RESULT (ic))->rematiCode = OP_SYMBOL (IC_RIGHT (ic))->rematiCode;
+        }
+
+      /* if cast to a generic pointer & the pointer being
+         cast is remat, then we can remat this cast as well */
+      if (ic->op == CAST &&
+        IS_SYMOP (IC_RIGHT (ic)) && OP_SYMBOL (IC_RIGHT (ic))->remat &&
+        !isOperandGlobal (IC_RESULT (ic)) && bitVectnBitsOn (OP_DEFS (IC_RESULT (ic))) == 1 && !IS_PARM (IC_RESULT (ic)) /* The receiving of the paramter is not accounted for in DEFS */)
+        {
+          sym_link *to_type = operandType (IC_LEFT (ic));
+          sym_link *from_type = operandType (IC_RIGHT (ic));
+          if (IS_GENPTR (to_type) && IS_PTR (from_type))
+            {
+              OP_SYMBOL (IC_RESULT (ic))->remat = 1;
+              OP_SYMBOL (IC_RESULT (ic))->rematiCode = ic;
+              OP_SYMBOL (IC_RESULT (ic))->usl.spillLoc = NULL;
+            }
+        }
+
+      /* In some cases redundant moves can be eliminated */
+      if (ic->op == GET_VALUE_AT_ADDRESS || ic->op == SET_VALUE_AT_ADDRESS ||
+        ic->op == IFX && operandSize (IC_COND (ic)) == 1 ||
+        ic->op == IPUSH && operandSize (IC_LEFT (ic)) == 1)
+        packRegsForOneuse (ic, &(IC_LEFT (ic)), ebp);
+    }
 }
-#endif
 
 /**
   Mark variables for assignment by the register allocator.
@@ -438,6 +607,8 @@ serialRegMark (eBBlock ** ebbs, int count)
 
   stm8_call_stack_size = 2; // Saving of register to stack temporarily.
 
+  D (D_ALLOC, ("serialRegMark for %s, currFunc->stack %d\n", currFunc->name, currFunc->stack));
+
   /* for all blocks */
   for (i = 0; i < count; i++)
     {
@@ -467,7 +638,7 @@ serialRegMark (eBBlock ** ebbs, int count)
 
           /* some don't need registers, since there is no result. */
           if (SKIP_IC2 (ic) ||
-              ic->op == JUMPTABLE || ic->op == IFX || ic->op == IPUSH || ic->op == IPOP || (IC_RESULT (ic) && POINTER_SET (ic)))
+              ic->op == JUMPTABLE || ic->op == IFX || ic->op == IPUSH || ic->op == IPOP || ic->op == SET_VALUE_AT_ADDRESS)
             continue;
 
           /* now we need to allocate registers only for the result */
@@ -475,7 +646,7 @@ serialRegMark (eBBlock ** ebbs, int count)
             {
               symbol *sym = OP_SYMBOL (IC_RESULT (ic));
 
-              D (D_ALLOC, ("serialRegAssign: in loop on result %p\n", sym));
+              D (D_ALLOC, ("serialRegAssign: in loop on result %p %s\n", sym, sym->name));
 
               if (sym->isspilt && sym->usl.spillLoc) // todo: Remove once remat is supported!
                 {
@@ -491,15 +662,22 @@ serialRegMark (eBBlock ** ebbs, int count)
                  or is already marked for the new allocator
                  or will not live beyond this instructions */
               if (!sym->nRegs ||
-                  sym->isspilt || sym->for_newralloc || sym->liveTo <= ic->seq)
+                  sym->isspilt || sym->for_newralloc || sym->liveTo <= ic->seq && (sym->nRegs <= 4 || ic->op != CALL && ic->op != PCALL))
                 {
-                  D (D_ALLOC, ("serialRegAssign: won't live long enough.\n"));
+                  D (D_ALLOC, ("serialRegMark: won't live long enough.\n"));
                   continue;
                 }
 
-              if (sym->nRegs > 4 && ic->op == CALL)
+              if (sym->usl.spillLoc && !sym->usl.spillLoc->_isparm) // I have no idea where these spill locations come from. Sometime two symbols even have the same spill location, whic tends to mess up stack allocation. THose that come from previous iterations in this loop would be okay, but those from outside are a problem.
                 {
-                  spillThis (sym, TRUE);
+                  sym->usl.spillLoc = 0;
+                  sym->isspilt = false;
+                }
+
+              if (sym->nRegs > 4 && ic->op == CALL) // To be allocated to stack due to the way long long return values are handled via a hidden pointer.
+                {
+                  sym->for_newralloc = 0;
+                  stm8SpillThis (sym, TRUE);
                 }
               else if (max_alloc_bytes >= sym->nRegs)
                 {
@@ -508,7 +686,7 @@ serialRegMark (eBBlock ** ebbs, int count)
                 }
               else if (!sym->for_newralloc)
                 {
-                  spillThis (sym, TRUE);
+                  stm8SpillThis (sym, TRUE);
                   printf ("Spilt %s due to byte limit.\n", sym->name);
                 }
             }
@@ -551,11 +729,11 @@ verifyRegsAssigned (operand * op, iCode * ic)
   if (completely_in_regs)
     return;
 
-  spillThis (sym, FALSE);
+  stm8SpillThis (sym, FALSE);
 }
 
-static void
-RegFix (eBBlock ** ebbs, int count)
+void
+stm8RegFix (eBBlock ** ebbs, int count)
 {
   int i;
 
@@ -597,23 +775,20 @@ void stm8_init_asmops (void);
 /* assignRegisters - assigns registers to each live range as need  */
 /*-----------------------------------------------------------------*/
 void
-stm8_assignRegisters (ebbIndex * ebbi)
+stm8_assignRegisters (ebbIndex *ebbi)
 {
   eBBlock **ebbs = ebbi->bbOrder;
   int count = ebbi->count;
   iCode *ic;
-#if 0 // ENABLE WHEN LINKER BUG #2198 IS FIXED
-  int i;
-#endif
 
   stm8_init_asmops();
 
+  transformPointerSet (ebbs, count);
+
   /* change assignments this will remove some
      live ranges reducing some register pressure */
-#if 0 // ENABLE WHEN LINKER BUG #2198 IS FIXED
-  for (i = 0; i < count; i++)
+  for (int i = 0; i < count; i++)
     packRegisters (ebbs[i]);
-#endif
 
   /* liveranges probably changed by register packing
      so we compute them again */
@@ -634,34 +809,20 @@ stm8_assignRegisters (ebbIndex * ebbi)
   /* Invoke optimal register allocator */
   ic = stm8_ralloc2_cc (ebbi);
 
-  /* Get spilllocs for all variables that have not been placed completely in regs */
-  RegFix (ebbs, count);
-
   /* redo the offsets for stacked automatic variables */
-  if (currFunc)
+  if (currFunc && !stm8_extend_stack && currFunc->stack + stm8_call_stack_size > 255)
     {
-      long b = currFunc->stack;
+      _G.slocNum = 0;
 
-      redoStackOffsets ();
+      /* Mark variables for assignment by the new allocator */
+      serialRegMark (ebbs, count);
 
-      /* Try again, using an extended stack this time. */
-      if (!stm8_extend_stack && currFunc->stack + stm8_call_stack_size > 255)
-        {
-          currFunc->stack = b;
+      stm8_extend_stack = TRUE;
 
-          /* Mark variables for assignment by the new allocator */
-          serialRegMark (ebbs, count);
+      /* Invoke optimal register allocator */
+      ic = stm8_ralloc2_cc (ebbi);
 
-          stm8_extend_stack = TRUE;
-
-          /* Invoke optimal register allocator */
-          ic = stm8_ralloc2_cc (ebbi);
-
-          /* Get spilllocs for all variables that have not been placed completely in regs */
-          RegFix (ebbs, count);
-
-          redoStackOffsets ();
-        }
+      //redoStackOffsets ();
     }
 
   if (options.dump_i_code)
@@ -671,5 +832,7 @@ stm8_assignRegisters (ebbIndex * ebbi)
     }
 
   genSTM8Code (ic);
+
+  _G.slocNum = 0;
 }
 
