@@ -37,7 +37,7 @@
 #define D(_s)
 #endif
 
-#define ISINST(l, i) (!STRNCASECMP((l), (i), sizeof(i) - 1))
+#define ISINST(l, i) (!STRNCASECMP((l), (i), sizeof(i) - 1) && (!(l)[sizeof(i) - 1] || isspace((unsigned char)((l)[sizeof(i) - 1]))))
 
 typedef enum
 {
@@ -204,11 +204,41 @@ findLabel (const lineNode *pl)
 /* Check if reading arg implies reading what. */
 static bool argCont(const char *arg, const char *what)
 {
+  wassert (arg);
+
   while(isspace (*arg) || *arg == ',')
     arg++;
+
+  if(arg[0] == '(' && arg[1] && arg[2] && (arg[2] != ')' && arg[3] != ')'))
+    return FALSE;
+
+  if(*arg == '(')
+    arg++;
+
   return(arg[0] == '#') ? FALSE : StrStr(arg, what) != NULL;
 }
 
+static bool
+z80MightBeParmInCallFromCurrentFunction(const char *what)
+{
+  if (strchr(what, 'l') && z80_regs_used_as_parms_in_calls_from_current_function[L_IDX])
+    return TRUE;
+  if (strchr(what, 'h') && z80_regs_used_as_parms_in_calls_from_current_function[H_IDX])
+    return TRUE;
+  if (strchr(what, 'e') && z80_regs_used_as_parms_in_calls_from_current_function[E_IDX])
+    return TRUE;
+  if (strchr(what, 'd') && z80_regs_used_as_parms_in_calls_from_current_function[D_IDX])
+    return TRUE;
+  if (strchr(what, 'c') && z80_regs_used_as_parms_in_calls_from_current_function[C_IDX])
+    return TRUE;
+  if (strchr(what, 'b') && z80_regs_used_as_parms_in_calls_from_current_function[B_IDX])
+    return TRUE;
+  if (strstr(what, "iy") && (z80_regs_used_as_parms_in_calls_from_current_function[IYL_IDX] || z80_regs_used_as_parms_in_calls_from_current_function[IYH_IDX]))
+    return TRUE;
+
+  return FALSE;
+}
+
 static bool
 z80MightRead(const lineNode *pl, const char *what)
 {
@@ -249,78 +279,74 @@ z80MightRead(const lineNode *pl, const char *what)
         return FALSE;
       }
       else // Fallback needed for calls through function pointers and for calls to literal addresses.
-      {
-        if (strchr(what, 'l') && z80_regs_used_as_parms_in_calls_from_current_function[L_IDX])
-          return TRUE;
-        if (strchr(what, 'h') && z80_regs_used_as_parms_in_calls_from_current_function[H_IDX])
-          return TRUE;
-        if (strchr(what, 'e') && z80_regs_used_as_parms_in_calls_from_current_function[E_IDX])
-          return TRUE;
-        if (strchr(what, 'd') && z80_regs_used_as_parms_in_calls_from_current_function[D_IDX])
-          return TRUE;
-        if (strchr(what, 'c') && z80_regs_used_as_parms_in_calls_from_current_function[C_IDX])
-          return TRUE;
-        if (strchr(what, 'b') && z80_regs_used_as_parms_in_calls_from_current_function[B_IDX])
-          return TRUE;
-        if (strstr(what, "iy") && (z80_regs_used_as_parms_in_calls_from_current_function[IYL_IDX] || z80_regs_used_as_parms_in_calls_from_current_function[IYH_IDX]))
-          return TRUE;
-        return FALSE;
-      }
+        return z80MightBeParmInCallFromCurrentFunction(what);
     }
 
-  if(ISINST(pl->line, "ret"))
-    return(isReturned(what));
+  if(ISINST(pl->line, "reti") || ISINST(pl->line, "retn"))
+    return(false);
+
+  if(ISINST(pl->line, "ret")) // --reserev-regs-iy uses ret in code gen for calls through function pointers
+    return(IY_RESERVED ? isReturned(what) || z80MightBeParmInCallFromCurrentFunction(what) : isReturned(what));
 
   if(!strcmp(pl->line, "ex\t(sp), hl") || !strcmp(pl->line, "ex\t(sp),hl"))
     return(!strcmp(what, "h") || !strcmp(what, "l"));
   if(!strcmp(pl->line, "ex\t(sp), ix") || !strcmp(pl->line, "ex\t(sp),ix"))
-    return(strstr(what, "ix"));
+    return(!!strstr(what, "ix"));
   if(!strcmp(pl->line, "ex\t(sp), iy") || !strcmp(pl->line, "ex\t(sp),iy"))
-    return(strstr(what, "iy"));
+    return(!!strstr(what, "iy"));
   if(!strcmp(pl->line, "ex\tde, hl") || !strcmp(pl->line, "ex\tde,hl"))
     return(!strcmp(what, "h") || !strcmp(what, "l") || !strcmp(what, "d") || !strcmp(what, "e"));
-  if(ISINST(pl->line, "ld\t"))
+  if(ISINST(pl->line, "ld"))
     {
-      if(strstr(strchr(pl->line, ','), what) && strchr(pl->line, ',')[1] != '#' && !(strchr(pl->line, ',')[1] == '(' && strchr(pl->line, ',')[2] == '#') && !(strchr(pl->line, ',')[1] == '(' && strchr(pl->line, ',')[3] != ')' && strchr(pl->line, ',')[4] != ')'))
-        return TRUE;
+      if(argCont(strchr(pl->line, ','), what))
+        return(true);
       if(*(strchr(pl->line, ',') - 1) == ')' && strstr(pl->line + 3, what) && (strchr(pl->line, '#') == 0 || strchr(pl->line, '#') > strchr(pl->line, ',')))
-        return TRUE;
-      return FALSE;
+        return(true);
+      return(false);
     }
 
   if(!strcmp(pl->line, "xor\ta, a") || !strcmp(pl->line, "xor\ta,a"))
-    return FALSE;
+    return(false);
 
-  if(ISINST(pl->line, "adc\t") ||
-     ISINST(pl->line, "add\t") ||
-     ISINST(pl->line, "and\t") ||
-     ISINST(pl->line, "sbc\t") ||
-     ISINST(pl->line, "sub\t") ||
-     ISINST(pl->line, "xor\t"))
+  if(ISINST(pl->line, "adc") ||
+     ISINST(pl->line, "add") ||
+     ISINST(pl->line, "and") ||
+     ISINST(pl->line, "sbc") ||
+     ISINST(pl->line, "sub") ||
+     ISINST(pl->line, "xor"))
     {
-      return(argCont(pl->line + 4, what));
+      const char *arg = pl->line + 4;
+      while(isspace(*arg))
+        arg++;
+      if(*arg == 'a' && *(arg + 1) == ',')
+        {
+          if(!strcmp(what, "a"))
+            return(true);
+          arg += 2;
+        }
+      return(argCont(arg, what));
     }
 
-  if(ISINST(pl->line, "or\t") || ISINST(pl->line, "cp\t") )
+  if(ISINST(pl->line, "or") || ISINST(pl->line, "cp") )
     {
       if(argCont(pl->line + 3, what))
-        return TRUE;
+        return(true);
       if(strcmp("a", what) == 0)
-        return TRUE;
-      return FALSE;
+        return(true);
+      return(false);
     }
 
   if(ISINST(pl->line, "neg"))
     return(strcmp(what, "a") == 0);
 
-  if(ISINST(pl->line, "pop\t"))
-    return FALSE;
+  if(ISINST(pl->line, "pop"))
+    return(false);
 
-  if(ISINST(pl->line, "push\t"))
+  if(ISINST(pl->line, "push"))
     return(strstr(pl->line + 5, what) != 0);
 
-  if(ISINST(pl->line, "dec\t") ||
-     ISINST(pl->line, "inc\t"))
+  if(ISINST(pl->line, "dec") ||
+     ISINST(pl->line, "inc"))
     {
       return(argCont(pl->line + 4, what));
     }
@@ -329,7 +355,7 @@ z80MightRead(const lineNode *pl, const char *what)
     return(!strcmp(what, "a"));
 
   if(ISINST(pl->line, "di") || ISINST(pl->line, "ei"))
-    return(FALSE);
+    return(false);
 
   // Rotate and shift group
   if(ISINST(pl->line, "rlca") ||
@@ -339,78 +365,77 @@ z80MightRead(const lineNode *pl, const char *what)
     {
       return(strcmp(what, "a") == 0);
     }
-  if(ISINST(pl->line, "rl\t") ||
-     ISINST(pl->line, "rr\t"))
+  if(ISINST(pl->line, "rl") ||
+     ISINST(pl->line, "rr"))
     {
       return(argCont(pl->line + 3, what));
     }
-  if(ISINST(pl->line, "rlc\t") ||
-     ISINST(pl->line, "sla\t") ||
-     ISINST(pl->line, "sra\t") ||
-     ISINST(pl->line, "srl\t"))
+  if(ISINST(pl->line, "rlc") ||
+     ISINST(pl->line, "sla") ||
+     ISINST(pl->line, "sra") ||
+     ISINST(pl->line, "srl"))
     {
       return(argCont(pl->line + 4, what));
     }
   if(!IS_GB && !IS_RAB &&
     (ISINST(pl->line, "rld") ||
      ISINST(pl->line, "rrd")))
-    return(strstr("ahl", what));
+    return(!!strstr("ahl", what));
 
   // Bit set, reset and test group
-  if(ISINST(pl->line, "bit\t") ||
-     ISINST(pl->line, "set\t") ||
-     ISINST(pl->line, "res\t"))
+  if(ISINST(pl->line, "bit") ||
+     ISINST(pl->line, "set") ||
+     ISINST(pl->line, "res"))
     {
       return(argCont(pl->line + 4, what));
     }
 
  if(ISINST(pl->line, "ccf") ||
     ISINST(pl->line, "nop"))
-    return FALSE;
+    return(false);
 
-  if(strncmp(pl->line, "jp\t", 3) == 0 ||
-     strncmp(pl->line, "jr\t", 3) == 0)
-    return FALSE;
+  if(ISINST(pl->line, "jp") || ISINST(pl->line, "jr"))
+    return(false);
 
-  if(ISINST(pl->line, "djnz\t"))
+  if(ISINST(pl->line, "djnz"))
     return(strchr(what, 'b') != 0);
 
   if(!IS_GB && ISINST(pl->line, "ldir"))
     return(!strcmp(what, "b") || !strcmp(what, "c") || !strcmp(what, "d") || !strcmp(what, "e") || !strcmp(what, "h") || !strcmp(what, "l"));
 
-  if(!IS_GB && !IS_RAB && ISINST(pl->line, "out\t"))
+  if(!IS_GB && !IS_RAB && ISINST(pl->line, "out"))
     return(strstr(strchr(pl->line + 4, ','), what) != 0 || strstr(pl->line + 4, "(c)") && (!strcmp(what, "b") || !strcmp(what, "c")));
-  if(!IS_GB && !IS_RAB && ISINST(pl->line, "in\t"))
+  if(!IS_GB && !IS_RAB && ISINST(pl->line, "in"))
     return(!strstr(strchr(pl->line + 4, ','), "(c)") && !strcmp(what, "a") || strstr(strchr(pl->line + 4, ','), "(c)") && (!strcmp(what, "b") || !strcmp(what, "c")));
 
   if(!IS_GB && !IS_RAB &&
-    (ISINST(pl->line, "ini\t") || ISINST(pl->line, "ind\t") || ISINST(pl->line, "inir\t") || ISINST(pl->line, "indr\t") ||
-    ISINST(pl->line, "outi\t") || ISINST(pl->line, "outd\t") || ISINST(pl->line, "otir\t") || ISINST(pl->line, "otdr\t")))
+    (ISINST(pl->line, "ini") || ISINST(pl->line, "ind") || ISINST(pl->line, "inir") || ISINST(pl->line, "indr") ||
+    ISINST(pl->line, "outi") || ISINST(pl->line, "outd") || ISINST(pl->line, "otir") || ISINST(pl->line, "otdr")))
     return(!strcmp(what, "b") || !strcmp(what, "c") || !strcmp(what, "h") || !strcmp(what, "l"));
 
-  if(IS_Z180 && ISINST(pl->line, "mlt\t"))
+  if((IS_Z180 || IS_EZ80_Z80) && ISINST(pl->line, "mlt"))
     return(argCont(pl->line + 4, what));
 
-  if(IS_Z180 && ISINST(pl->line, "tst\t"))
+  if((IS_Z180 || IS_EZ80_Z80) && ISINST(pl->line, "tst"))
     return(argCont(pl->line + 4, what));
 
   if(IS_RAB && ISINST(pl->line, "mul"))
     return(!strcmp(what, "b") || !strcmp(what, "c") || !strcmp(what, "d") || !strcmp(what, "e"));
 
-  if(IS_RAB && ISINST(pl->line, "bool\t"))
+  if(IS_RAB && ISINST(pl->line, "bool"))
     return(argCont(pl->line + 5, what));
 
   /* TODO: Can we know anything about rst? */
   if(ISINST(pl->line, "rst"))
-    return(TRUE);
+    return(true);
 
-  return TRUE;
+  return(true);
 }
 
 static bool
 z80UncondJump(const lineNode *pl)
 {
-  if((ISINST(pl->line, "jp\t") || ISINST(pl->line, "jr\t")) &&
+  if((ISINST(pl->line, "jp") || ISINST(pl->line, "jr")) &&
      strchr(pl->line, ',') == 0)
     return TRUE;
   return FALSE;
@@ -419,9 +444,9 @@ z80UncondJump(const lineNode *pl)
 static bool
 z80CondJump(const lineNode *pl)
 {
-  if(((ISINST(pl->line, "jp\t") || ISINST(pl->line, "jr\t")) &&
+  if(((ISINST(pl->line, "jp") || ISINST(pl->line, "jr")) &&
       strchr(pl->line, ',') != 0) ||
-     ISINST(pl->line, "djnz\t"))
+     ISINST(pl->line, "djnz"))
     return TRUE;
   return FALSE;
 }
@@ -434,26 +459,26 @@ z80SurelyWrites(const lineNode *pl, const char *what)
   if(strcmp(what, "ixl") == 0 || strcmp(what, "ixh") == 0)
     what = "ix";
 
-  if(ISINST(pl->line, "xor\t") && strcmp(what, "a") == 0)
-    return TRUE;
-  if(ISINST(pl->line, "ld\t") && strncmp(pl->line + 3, "hl", 2) == 0 && (what[0] == 'h' || what[0] == 'l'))
-    return TRUE;
-  if(ISINST(pl->line, "ld\t") && strncmp(pl->line + 3, "de", 2) == 0 && (what[0] == 'd' || what[0] == 'e'))
-    return TRUE;
-  if(ISINST(pl->line, "ld\t") && strncmp(pl->line + 3, "bc", 2) == 0 && (what[0] == 'b' || what[0] == 'c'))
-    return TRUE;
-  if((ISINST(pl->line, "ld\t") || ISINST(pl->line, "in\t"))
+  if(ISINST(pl->line, "xor") && strcmp(what, "a") == 0)
+    return(true);
+  if(ISINST(pl->line, "ld") && strncmp(pl->line + 3, "hl", 2) == 0 && (what[0] == 'h' || what[0] == 'l'))
+    return(true);
+  if(ISINST(pl->line, "ld") && strncmp(pl->line + 3, "de", 2) == 0 && (what[0] == 'd' || what[0] == 'e'))
+    return(true);
+  if(ISINST(pl->line, "ld") && strncmp(pl->line + 3, "bc", 2) == 0 && (what[0] == 'b' || what[0] == 'c'))
+    return(true);
+  if((ISINST(pl->line, "ld") || ISINST(pl->line, "in"))
     && strncmp(pl->line + 3, what, strlen(what)) == 0 && pl->line[3 + strlen(what)] == ',')
-    return TRUE;
-  if(ISINST(pl->line, "pop\t") && strstr(pl->line + 4, what))
-    return TRUE;
-  if(ISINST(pl->line, "call\t") && strchr(pl->line, ',') == 0)
+    return(true);
+  if(ISINST(pl->line, "pop") && strstr(pl->line + 4, what))
+    return(true);
+  if(ISINST(pl->line, "call") && strchr(pl->line, ',') == 0)
     {
       const symbol *f = findSym (SymbolTab, 0, pl->line + 6);
       const bool *preserved_regs;
 
       if(!strcmp(what, "ix"))
-        return FALSE;
+        return(false);
 
       if(f)
           preserved_regs = f->type->funcAttrs.preserved_regs;
@@ -476,11 +501,26 @@ z80SurelyWrites(const lineNode *pl, const char *what)
         return !preserved_regs[IYL_IDX] && !preserved_regs[IYH_IDX];
     }
   if(strcmp(pl->line, "ret") == 0)
-    return TRUE;
-  if(ISINST(pl->line, "ld\tiy") && strncmp(what, "iy", 2) == 0)
-    return TRUE;
+    return true;
+  if(strcmp(pl->line, "ld\tiy")  == 0 && strncmp(what, "iy", 2) == 0)
+    return true;
 
-  return FALSE;
+  if (IS_Z180 || IS_EZ80_Z80)
+    {
+      if (ISINST(pl->line, "mlt"))
+        return(strchr(pl->line + 4, *what) != 0);
+
+      if (ISINST(pl->line, "otim") ||
+        ISINST(pl->line, "otimr") ||
+        ISINST(pl->line, "otdm") ||
+        ISINST(pl->line, "otdmr"))
+        return(strchr("bchl", *what) != NULL);
+
+      if (ISINST(pl->line, "in0"))
+        return(!strncmp(pl->line + 4, what, strlen(what)));
+    }
+
+  return(false);
 }
 
 static bool
@@ -837,23 +877,37 @@ int z80instructionSize(lineNode *pl)
     return(6);
 
   /* All ld instructions */
-  if(ISINST(pl->line, "ld\t") || ISINST(pl->line, "ld "))
+  if(ISINST(pl->line, "ld"))
     {
-      /* These 3 are the only cases of 4 byte long ld instructions. */
+      /* These 4 are the only cases of 4 byte long ld instructions. */
       if(!STRNCASECMP(op1start, "ix", 2) || !STRNCASECMP(op1start, "iy", 2))
         return(4);
       if((argCont(op1start, "(ix)") || argCont(op1start, "(iy)")) && op2start[0] == '#')
         return(4);
+
       if(op1start[0] == '('               && STRNCASECMP(op1start, "(bc)", 4) &&
          STRNCASECMP(op1start, "(de)", 4) && STRNCASECMP(op1start, "(hl)", 4) &&
-         STRNCASECMP(op2start, "hl", 2)   && STRNCASECMP(op2start, "a", 1))
+         STRNCASECMP(op2start, "hl", 2)   && STRNCASECMP(op2start, "a", 1) ||
+         op2start[0] == '('               && STRNCASECMP(op2start, "(bc)", 4) &&
+         STRNCASECMP(op1start, "(de)", 4) && STRNCASECMP(op2start, "(hl)", 4) &&
+         STRNCASECMP(op1start, "hl", 2)   && STRNCASECMP(op1start, "a", 1))
         return(4);
 
+      /* Rabbit 16-bit pointer load */
       if(IS_RAB && !STRNCASECMP(op1start, "hl", 2) && (argCont(op2start, "(hl)") || argCont(op2start, "(iy)")))
         return(4);
       if(IS_RAB && !STRNCASECMP(op1start, "hl", 2) && (argCont(op2start, "(sp)") || argCont(op2start, "(ix)")))
         return(3);
 
+      if(IS_EZ80_Z80 && /* eZ80 16-bit pointer load */
+        (!STRNCASECMP(op1start, "bc", 2) || !STRNCASECMP(op1start, "de", 2) || !STRNCASECMP(op1start, "hl", 2) || !STRNCASECMP(op1start, "ix", 2) || !STRNCASECMP(op1start, "iy", 2)))
+        {
+          if (!STRNCASECMP(op2start, "(hl)", 4))
+            return(2);
+          if (argCont(op2start, "(ix)") || argCont(op2start, "(iy)"))
+            return(3);
+        }
+
       /* These 4 are the only remaining cases of 3 byte long ld instructions. */
       if(argCont(op2start, "(ix)") || argCont(op2start, "(iy)"))
         return(3);
@@ -1008,13 +1062,16 @@ int z80instructionSize(lineNode *pl)
   if(ISINST(pl->line, "in") || ISINST(pl->line, "out") || ISINST(pl->line, "ot"))
     return(2);
 
+  if((IS_Z180 || IS_EZ80_Z80) && (ISINST(pl->line, "in0") || ISINST(pl->line, "out0")))
+    return(3);
+
   if(ISINST(pl->line, "di") || ISINST(pl->line, "ei"))
     return(1);
 
-  if(IS_Z180 && ISINST(pl->line, "mlt"))
+  if((IS_Z180 || IS_EZ80_Z80) && ISINST(pl->line, "mlt"))
     return(2);
 
-  if(IS_Z180 && ISINST(pl->line, "tst"))
+  if((IS_Z180 || IS_EZ80_Z80) && ISINST(pl->line, "tst"))
     return((op1start[0] == '#' || op2start && op1start[0] == '#') ? 3 : 2);
   
   if(IS_RAB && ISINST(pl->line, "mul"))
@@ -1034,18 +1091,31 @@ int z80instructionSize(lineNode *pl)
 
   if(IS_RAB && ISINST(pl->line, "bool"))
     return(!STRNCASECMP(op1start, "hl", 2) ? 1 : 2);
-  
-  if(ISINST(pl->line, ".db"))
+
+  if(IS_EZ80_Z80 && (ISINST(pl->line, "lea") || ISINST(pl->line, "pea")))
+    return(3);
+
+  if(ISINST(pl->line, ".db") || ISINST(pl->line, ".byte"))
     {
       int i, j;
       for(i = 1, j = 0; pl->line[j]; i += pl->line[j] == ',', j++);
       return(i);
     }
 
+  if(ISINST(pl->line, ".dw") || ISINST(pl->line, ".word"))
+    {
+      int i, j;
+      for(i = 1, j = 0; pl->line[j]; i += pl->line[j] == ',', j++);
+      return(i * 2);
+    }
+
   /* If the instruction is unrecognized, we shouldn't try to optimize.  */
   /* For all we know it might be some .ds or similar possibly long line */
   /* Return a large value to discourage optimization.                   */
-  werrorfl(pl->ic->filename, pl->ic->lineno, W_UNRECOGNIZED_ASM, __FUNCTION__, 999, pl->line);
+  if (pl->ic)
+    werrorfl(pl->ic->filename, pl->ic->lineno, W_UNRECOGNIZED_ASM, __func__, 999, pl->line);
+  else
+    werrorfl("unknown", 0, W_UNRECOGNIZED_ASM, __func__, 999, pl->line);
   return(999);
 }
 
