@@ -13,8 +13,8 @@
 #endif
 
 /* PR 6534 */
-/* GCSE unified the two i<0 tests, but if-conversion to ui=abs(i) 
-   insertted the code at the wrong place corrupting the i<0 test.  */
+/* GCSE unified the two i<0 tests, but if-conversion to ui=abs(i)
+   inserted the code at the wrong place corrupting the i<0 test.  */
 
 static char *
 inttostr (long i, char buf[N])
@@ -35,6 +35,7 @@ inttostr (long i, char buf[N])
 void
 testTortureExecute (void)
 {
+#ifndef __SDCC_pdk14 // Lack of memory
   char buf[N], *p;
 
   p = inttostr (-1, buf);
@@ -45,4 +46,5 @@ testTortureExecute (void)
   ASSERT(p == buf + (N-1) - 2);
 
   return;
+#endif
 }
