@@ -8,8 +8,7 @@
 #pragma std_c99
 #endif
 
-// Todo: Enable when sdcc supports long long modulo!
-#if !defined (__SDCC_mcs51) && !defined (__SDCC_ds390) && !defined (__SDCC_hc08) && !defined (__SDCC_s08)
+#if !defined (__SDCC_pic14) && !defined(__SDCC_pdk14) // Lack of memory
 /* PR c/19606
    The C front end used to shorten the type of a division to a type
    that does not preserve the semantics of the original computation.
@@ -32,7 +31,7 @@ bar (void)
 void
 testTortureExecute (void)
 {
-#if !defined (__SDCC_mcs51) && !defined (__SDCC_ds390) && !defined (__SDCC_hc08) && !defined (__SDCC_s08)
+#if !defined (__SDCC_pic14) && !defined(__SDCC_pdk14) // Lack of memory
 #if !defined (PORT_HOST) // failed in test-host
   int r;
 
